# -*- coding: utf-8 -*-
from __future__ import print_function
import fnmatch
import os
import re
from abjad.tools import stringtools
from abjad.tools.commandlinetools.CommandlineScript import CommandlineScript
try:
    input = raw_input
except NameError:
    pass


class ReplaceScript(CommandlineScript):
    r'''Replaces text in files recursively.

    ..  shell::

        ajv replace text --help

    Multiple patterns for excluding files or folders can be specified by
    restating the `--without-files` or `--without-dirs` commands:

    ..  code-block:: bash

        abjad$ ajv replace text . foo bar -F *.txt -F *.rst -F *.htm

    '''

    ### CLASS VARIABLES ###

    alias = 'replace'
    short_description = 'Replace text.'

    ### PRIVATE METHODS ###

    def _get_naive_search_callable(self, arguments):

        class NaiveSearch(object):

            def __init__(self, pattern):
                self.pattern = pattern

            def __call__(self, line, pos):
                index = line.find(self.pattern, pos)
                if 0 <= index:
                    return index, len(self.pattern)
                return -1, 0

        return NaiveSearch(arguments.old)

    def _get_regex_search_callable(self, arguments):

        class RegexSearch(object):

            def __init__(self, pattern, escaped=False, whole_words_only=False):
                try:
                    if escaped:
                        pattern = re.escape(pattern)
                    if whole_words_only:
                        pattern += r'\b'
                    self.pattern = re.compile(pattern)
                    self.whole_words_only = whole_words_only
                except:
                    message = "can't compile {!r} as a regex pattern."
                    message = message.format(pattern)
                    raise ValueError(message)

            def __call__(self, line, pos):
                start, length = self._search(line, pos)
                if self.whole_words_only and 0 < start:
                    while start != -1 and (
                        line[start - 1].isalnum() or line[start - 1] == '_'):
                        start, length = self._search(line, start + length)
                return start, length

            def _search(self, line, pos):
                match = self.pattern.search(line, pos)
                if match is None:
                    return -1, 0
                return match.start(), match.end() - match.start()

        return RegexSearch(
            arguments.old,
            escaped=not arguments.regex,
            whole_words_only=arguments.whole_words_only,
            )

    def _process_args(self, arguments):
        print('Replacing {!r} with {!r} ...'.format(arguments.old, arguments.new))
        skipped_dirs_patterns = self.skipped_directories + arguments.without_dirs
        skipped_files_patterns = self.skipped_files + arguments.without_files
        if arguments.regex or (not arguments.regex and arguments.whole_words_only):
            arguments.old = self._get_regex_search_callable(arguments)
            index, length = arguments.old('', 0)
            if 0 <= index:
                message = 'regex pattern {!r} matches the empty string.'
                message = message.format(arguments.old.pattern.pattern)
                raise ValueError(message)
        else:
            arguments.old = self._get_naive_search_callable(arguments)
        changed_file_count = 0
        changed_line_count = 0
        changed_item_count = 0
        for root, dirs, files in os.walk(arguments.path):
            dirs_to_remove = []
            for dir in dirs:
                for pattern in skipped_dirs_patterns:
                    if fnmatch.fnmatch(dir, pattern):
                        dirs_to_remove.append(dir)
                        break
            for dir in dirs_to_remove:
                dirs.remove(dir)
            for file in sorted(files):
                valid = True
                for pattern in skipped_files_patterns:
                    if fnmatch.fnmatch(file, pattern):
                        valid = False
                        break
                if not valid:
                    continue
                changed_lines, changed_items = self._process_file(
                    arguments, os.path.join(root, file))
                if changed_lines:
                    changed_file_count += 1
                    changed_line_count += changed_lines
                    changed_item_count += changed_items
        print()
        item_identifier = stringtools.pluralize('instance', changed_item_count)
        line_identifier = stringtools.pluralize('line', changed_line_count)
        file_identifier = stringtools.pluralize('file', changed_file_count)
        message = '\tReplaced {} {} over {} {} in {} {}.'
        message = message.format(
            changed_item_count,
            item_identifier,
            changed_line_count,
            line_identifier,
            changed_file_count,
            file_identifier,
            )
        print(message)

    def _process_file(self, arguments, path):
        changed_items = 0
        changed_lines = 0
        try:
            with open(path, 'r') as f:
                lines = f.read().split('\n')
        except UnicodeDecodeError:
            print('FAILED READING: {}'.format(path))
            return changed_lines, changed_items
        results = []
        for i, line in enumerate(lines):
            line, changes = self._process_line(
                line, i, path, arguments.old, arguments.new, arguments.force, arguments.verbose)
            results.append(line)
            if changes:
                changed_items += changes
                changed_lines += 1
        if results != lines:
            with open(path, 'w') as f:
                f.write('\n'.join(results))
        return changed_lines, changed_items

    def _process_line(
        self,
        line,
        line_number,
        file_name,
        search,
        replacement,
        force,
        verbose,
        ):
        index, changes = 0, 0
        index, length = search(line, index)
        while 0 <= index:
            should_replace = False
            replaced_line = line[:index] + replacement + line[index + length:]
            carats = (' ' * index) + ('^' * length)
            if force:
                should_replace = True
                if verbose:
                    print()
                    print('{}: {}'.format(file_name, line_number))
                    print('-{}'.format(line))
                    print('+{}'.format(replaced_line))
            else:
                print()
                print('{}: {}'.format(file_name, line_number))
                print()
                print('{}'.format(line))
                print('{}'.format(carats))
                print()
                result = input('Replace? [Y/n] > ').lower()
                while result not in ('', 'y', 'yes', 'n', 'no'):
                    result = input('Replace? [Y/n] > ').lower()
                if result in ('', 'y', 'yes'):
                    should_replace = True
            if should_replace:
                index += length - (length - len(replacement))
                line = replaced_line
                changes += 1
            else:
                index += length
            index, length = search(line, index)
        return line, changes

    def _setup_argument_parser(self, parser):
        parser.add_argument(
            'path',
            default=os.getcwd(),
            help='directory tree to be recursed over',
            nargs='?',
            type=self._validate_path,
            )
        parser.add_argument(
            'old',
            help='old text',
            )
        parser.add_argument(
            'new',
            help='new text',
            )
        parser.add_argument(
            '--verbose',
            action='store_true',
            help='print replacement info even when --force flag is set.',
            )
        parser.add_argument(
            '-Y', '--force',
            action='store_true',
            help='force "yes" to every replacement'
            )
        parser.add_argument(
            '-R', '--regex',
            action='store_true',
            help='treat "old" as a regular expression',
            )
        parser.add_argument(
            '-W', '--whole-words-only',
            action='store_true',
            help='''match only whole words, similar to grep's "-w" flag''',
            )
        parser.add_argument(
            '-F', '--without-files',
            action='append',
            default=[],
            help='Exclude files matching pattern(s)',
            metavar='PATTERN',
            )
        parser.add_argument(
            '-D', '--without-dirs',
            action='append',
            default=[],
            help='Exclude folders matching pattern(s)',
            metavar='PATTERN',
            )

    ### PUBLIC PROPERTIES ###

    @property
    def skipped_directories(self):
        r'''Skipped directories.

        Returns list.
        '''
        return [
            '.svn',
            '.git',
            'build'
            ]

    @property
    def skipped_files(self):
        r'''Skipped files.

        Returns list.
        '''
        return [
            __file__,
            self.program_name,
            '*.backup',
            '*.doctree',
            '*.gif',
            '*.jpg',
            '*.jpeg',
            '*.nb',
            '*.pdf',
            '*.pickle',
            '*.pkl',
            '*.png',
            '*.ps',
            '*.pyc',
            '*.rtf',
            '*.txt',
            '.DS_Store',
            ]



# -*- coding: utf-8 -*-
import pytest
from abjad import *


def test_scoretools_Score___setattr___01():
    r'''Slots constrain score attributes.
    '''

    score = Score([])

    assert pytest.raises(AttributeError, "score.foo = 'bar'")



# -*- coding: utf-8 -*-
from abjad import *


def test_quantizationtools_BeatwiseQSchema___init___01():

    item_a = quantizationtools.BeatwiseQSchemaItem(
        search_tree=quantizationtools.UnweightedSearchTree({2: None}))
    item_b = quantizationtools.BeatwiseQSchemaItem(tempo=((1, 4), 76))
    item_c = quantizationtools.BeatwiseQSchemaItem(beatspan=(1, 8))

    schema = quantizationtools.BeatwiseQSchema(
        {2: item_a, 4: item_b, 7: item_c},
        beatspan=durationtools.Duration(1, 32),
        search_tree=quantizationtools.UnweightedSearchTree({3: None}),
        tempo=Tempo((1, 16), 32)
        )

    assert len(schema.items) == 3
    assert schema.beatspan == durationtools.Duration(1, 32)
    assert schema.search_tree == quantizationtools.UnweightedSearchTree({3: None})
    assert schema.tempo == Tempo((1, 16), 32)


def test_quantizationtools_BeatwiseQSchema___init___02():

    schema = quantizationtools.BeatwiseQSchema()

    assert len(schema.items) == 0
    assert schema.beatspan == durationtools.Duration(1, 4)
    assert schema.search_tree == quantizationtools.UnweightedSearchTree()
    assert schema.tempo == Tempo((1, 4), 60)



# -*- coding: utf-8 -*-
from abjad import graphtools
from abjad import stringtools


def test_documentationtools_GraphvizGraph_01():

    graph = graphtools.GraphvizGraph()
    graph.node_attributes['shape'] = 'record'

    struct_1 = graphtools.GraphvizNode()
    struct_1_field_0 = graphtools.GraphvizField(label='left')
    struct_1.append(struct_1_field_0)
    struct_1_field_1 = graphtools.GraphvizField(label='middle')
    struct_1.append(struct_1_field_1)
    struct_1_field_2 = graphtools.GraphvizField(label='right')
    struct_1.append(struct_1_field_2)

    struct_2 = graphtools.GraphvizNode()
    struct_2_field_0 = graphtools.GraphvizField(label='one')
    struct_2.append(struct_2_field_0)
    struct_2_field_1 = graphtools.GraphvizField(label='two')
    struct_2.append(struct_2_field_1)

    struct_3 = graphtools.GraphvizNode()
    struct_3_field_hello = graphtools.GraphvizField(
        label='hello')
    struct_3.append(struct_3_field_hello)
    outer_group = graphtools.GraphvizGroup()
    struct_3.append(outer_group)
    struct_3_field_b = graphtools.GraphvizField(label='b')
    outer_group.append(struct_3_field_b)
    inner_group = graphtools.GraphvizGroup()
    outer_group.append(inner_group)
    struct_3_field_c = graphtools.GraphvizField(label='c')
    inner_group.append(struct_3_field_c)
    struct_3_field_d = graphtools.GraphvizField(label='d')
    inner_group.append(struct_3_field_d)
    struct_3_field_e = graphtools.GraphvizField(label='e')
    inner_group.append(struct_3_field_e)
    struct_3_field_f = graphtools.GraphvizField(label='f')
    outer_group.append(struct_3_field_f)
    struct_3_field_g = graphtools.GraphvizField(label='g')
    struct_3.append(struct_3_field_g)
    struct_3_field_h = graphtools.GraphvizField(label='h')
    struct_3.append(struct_3_field_h)

    graph.extend((struct_1, struct_2, struct_3))

    graphtools.GraphvizEdge().attach(
        struct_1_field_1,
        struct_2_field_0,
        )
    graphtools.GraphvizEdge().attach(
        struct_1_field_2,
        struct_3_field_d,
        )

    graphviz_format = str(graph)

    assert graphviz_format == stringtools.normalize(
        '''
        digraph G {
            node [shape=record];
            node_0 [label="<f_0> left | <f_1> middle | <f_2> right"];
            node_1 [label="<f_0> one | <f_1> two"];
            node_2 [label="<f_0> hello | { <f_1_0> b | { <f_1_1_0> c | <f_1_1_1> d | <f_1_1_2> e } | <f_1_2> f } | <f_2> g | <f_3> h"];
            node_0:f_1 -> node_1:f_0;
            node_0:f_2 -> node_2:f_1_1_1;
        }
        '''
        )



# -*- coding: utf-8 -*-


def extend_measures_and_apply_full_measure_tuplets(argument, supplement):
    r'''Extends measures in `argument` with `supplement` and applies
    full-measure tuplets to contents of measures.

    ..  container:: example

        ::

            >>> staff = Staff([Measure((2, 8), "c'8 d'8"), Measure((3, 8), "e'8 f'8 g'8")])

        ..  doctest::

            >>> f(staff)
            \new Staff {
                {
                    \time 2/8
                    c'8
                    d'8
                }
                {
                    \time 3/8
                    e'8
                    f'8
                    g'8
                }
            }

        ::

            >>> supplement = [Rest((1, 16))]
            >>> scoretools.extend_measures_and_apply_full_measure_tuplets(
            ...     staff,
            ...     supplement,
            ...     )

        ..  doctest::

            >>> f(staff)
            \new Staff {
                {
                    \time 2/8
                    \times 4/5 {
                        c'8
                        d'8
                        r16
                    }
                }
                {
                    \time 3/8
                    \tweak text #tuplet-number::calc-fraction-text
                    \times 6/7 {
                        e'8
                        f'8
                        g'8
                        r16
                    }
                }
            }

    Returns none.
    '''
    from abjad.tools import scoretools

    return scoretools.apply_full_measure_tuplets_to_contents_of_measures(
        argument, supplement)



# -*- coding: utf-8 -*-
from abjad.tools import datastructuretools
from abjad.tools import mathtools
from abjad.tools import scoretools
from abjad.tools import sequencetools
from abjad.tools.abctools import AbjadValueObject


class BurnishSpecifier(AbjadValueObject):
    r'''Burnish specifier.

    ..  container:: example

        Forces first leaf of each division to be a rest:

        ::

            >>> burnish_specifier = rhythmmakertools.BurnishSpecifier(
            ...     left_classes=[Rest],
            ...     left_counts=[1],
            ...     )

    ..  container:: example

        Forces the first three leaves of each division to be rests:

        ::

            >>> burnish_specifier = rhythmmakertools.BurnishSpecifier(
            ...     left_classes=[Rest],
            ...     left_counts=[3],
            ...     )

    ..  container:: example

        Forces last leaf of each division to be a rest:

        ::

            >>> burnish_specifier = rhythmmakertools.BurnishSpecifier(
            ...     right_classes=[Rest],
            ...     right_counts=[1],
            ...     )

    ..  container:: example

        Forces the last three leaves of each division to be rests:

        ::

            >>> burnish_specifier = rhythmmakertools.BurnishSpecifier(
            ...     right_classes=[Rest],
            ...     right_counts=[3],
            ...     )

    ..  container:: example

        Forces the first leaf of every even-numbered division to be a rest;
        forces the first leaf of every odd-numbered division to be a note.

        ::

            >>> burnish_specifier = rhythmmakertools.BurnishSpecifier(
            ...     left_classes=[Rest, Note],
            ...     left_counts=[1],
            ...     )

    ..  container:: example

        Forces the last leaf of every even-numbered division to be a rest;
        forces the last leaf of every odd-numbered division to be a note.

        ::

            >>> burnish_specifier = rhythmmakertools.BurnishSpecifier(
            ...     right_classes=[Rest, Note],
            ...     right_counts=[1],
            ...     )

    ..  container:: example

        Forces the first leaf of every even-numbered division to be a rest;
        leave the first leaf of every odd-numbered division unchanged.

        ::

            >>> burnish_specifier = rhythmmakertools.BurnishSpecifier(
            ...     left_classes=[Rest, 0],
            ...     left_counts=[1],
            ...     )

    ..  container:: example

        Forces the last leaf of every even-numbered division to be a rest;
        leave the last leaf of every odd-numbered division unchanged.

        ::

            >>> burnish_specifier = rhythmmakertools.BurnishSpecifier(
            ...     right_classes=[Rest, 0],
            ...     right_counts=[1],
            ...     )

    Burnish specifiers are immutable.
    '''

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Specifiers'

    __slots__ = (
        '_left_classes',
        '_left_counts',
        '_middle_classes',
        '_outer_divisions_only',
        '_right_counts',
        '_right_classes',
        )

    _publish_storage_format = True

    ### INITIALIZER ###

    def __init__(
        self,
        left_classes=None,
        middle_classes=None,
        right_classes=None,
        left_counts=None,
        right_counts=None,
        outer_divisions_only=False,
        ):
        assert isinstance(outer_divisions_only, bool)
        self._outer_divisions_only = outer_divisions_only
        if left_classes is not None:
            left_classes = tuple(left_classes)
        if middle_classes is not None:
            middle_classes = tuple(middle_classes)
        if middle_classes == (0,):
            middle_classes = ()
        if right_classes is not None:
            right_classes = tuple(right_classes)
        if left_counts is not None:
            left_counts = tuple(left_counts)
        if right_counts is not None:
            right_counts = tuple(right_counts)
        assert self._is_sign_tuple(left_classes)
        assert self._is_sign_tuple(middle_classes)
        assert self._is_sign_tuple(right_classes)
        assert self._is_length_tuple(left_counts)
        assert self._is_length_tuple(right_counts)
        self._left_classes = left_classes
        self._middle_classes = middle_classes
        self._right_classes = right_classes
        self._left_counts = left_counts
        self._right_counts = right_counts

    ### SPECIAL METHODS ###

    def __call__(self, divisions, helper_functions=None, rotation=None):
        r'''Calls burnish specifier on `divisions`.

        Returns list of burnished divisions.
        '''
        input_ = self._rotate_input(helper_functions=None, rotation=None)
        if self.outer_divisions_only:
            return self._burnish_outer_divisions(input_, divisions)
        else:
            return self._burnish_each_division(input_, divisions)

    def __format__(self, format_specification=''):
        r'''Formats burnish specifier.

        ..  container:: example

            ::

                >>> burnish_specifier = rhythmmakertools.BurnishSpecifier(
                ...     left_classes=[Rest, 0],
                ...     left_counts=[1],
                ...     )

            ::

                >>> f(burnish_specifier)
                rhythmmakertools.BurnishSpecifier(
                    left_classes=[
                        scoretools.Rest,
                        0,
                        ],
                    left_counts=[1],
                    )

        Returns string.
        '''
        return AbjadValueObject.__format__(
            self,
            format_specification=format_specification,
            )

    def __repr__(self):
        r'''Gets interpreter representation.

        ..  container:: example

            ::

                >>> burnish_specifier = rhythmmakertools.BurnishSpecifier(
                ...     left_classes=[Rest, 0],
                ...     left_counts=[1],
                ...     )

            ::

                >>> burnish_specifier
                BurnishSpecifier(left_classes=[Rest, 0], left_counts=[1])

        Returns string.
        '''
        return super(BurnishSpecifier, self).__repr__()

    ### PRIVATE METHODS ###

    @staticmethod
    def _burnish_division_part(division_part, token):
        assert len(division_part) == len(token)
        new_division_part = []
        for number, i in zip(division_part, token):
            if i in (-1, scoretools.Rest):
                new_division_part.append(-abs(number))
            elif i == 0:
                new_division_part.append(number)
            elif i in (1, scoretools.Note):
                new_division_part.append(abs(number))
            else:
                raise ValueError
        new_division_part = type(division_part)(new_division_part)
        return new_division_part

    @classmethod
    def _burnish_each_division(class_, input_, divisions):
        left_classes = input_['left_classes']
        middle_classes = input_['middle_classes']
        right_classes = input_['right_classes']
        left_counts = input_['left_counts']
        left_counts = left_counts or datastructuretools.CyclicTuple([0])
        right_counts = input_['right_counts']
        right_counts = right_counts or datastructuretools.CyclicTuple([0])
        lefts_index, rights_index = 0, 0
        burnished_divisions = []
        for division_index, division in enumerate(divisions):
            left_count = left_counts[division_index]
            left = left_classes[lefts_index:lefts_index + left_count]
            lefts_index += left_count
            right_count = right_counts[division_index]
            right = right_classes[rights_index:rights_index + right_count]
            rights_index += right_count
            available_left_count = len(division)
            left_count = min([left_count, available_left_count])
            available_right_count = len(division) - left_count
            right_count = min([right_count, available_right_count])
            middle_count = len(division) - left_count - right_count
            left = left[:left_count]
            if middle_classes:
                middle = middle_count * [middle_classes[division_index]]
            else:
                middle = middle_count * [0]
            right = right[:right_count]
            left_part, middle_part, right_part = sequencetools.Sequence(
                    division).partition_by_counts(
                    [left_count, middle_count, right_count],
                    cyclic=False,
                    overhang=False,
                    )
            left_part = class_._burnish_division_part(left_part, left)
            middle_part = class_._burnish_division_part(middle_part, middle)
            right_part = class_._burnish_division_part(right_part, right)
            burnished_division = left_part + middle_part + right_part
            burnished_divisions.append(burnished_division)
        unburnished_weights = [mathtools.weight(x) for x in divisions]
        burnished_weights = [mathtools.weight(x) for x in burnished_divisions]
        assert burnished_weights == unburnished_weights
        return burnished_divisions

    @classmethod
    def _burnish_outer_divisions(class_, input_, divisions):
        for list_ in divisions:
            assert all(isinstance(_, int) for _ in list_), repr(list_)
        left_classes = input_['left_classes']
        middle_classes = input_['middle_classes']
        right_classes = input_['right_classes']
        left_counts = input_['left_counts']
        left_counts = left_counts or datastructuretools.CyclicTuple([0])
        right_counts = input_['right_counts']
        right_counts = right_counts or datastructuretools.CyclicTuple([0])
        burnished_divisions = []
        left_count = 0
        if left_counts:
            left_count = left_counts[0]
        left = left_classes[:left_count]
        right_count = 0
        if right_counts:
            right_count = right_counts[0]
        right = right_classes[:right_count]
        if len(divisions) == 1:
            available_left_count = len(divisions[0])
            left_count = min([left_count, available_left_count])
            available_right_count = len(divisions[0]) - left_count
            right_count = min([right_count, available_right_count])
            middle_count = len(divisions[0]) - left_count - right_count
            left = left[:left_count]
            if not middle_classes:
                middle_classes = [1]
            middle = [middle_classes[0]]
            middle = middle_count * middle
            right = right[:right_count]
            left_part, middle_part, right_part = sequencetools.Sequence(
                    divisions[0]).partition_by_counts(
                    [left_count, middle_count, right_count],
                    cyclic=False,
                    overhang=Exact,
                    )
            left_part = class_._burnish_division_part(left_part, left)
            middle_part = class_._burnish_division_part(middle_part, middle)
            right_part = class_._burnish_division_part(right_part, right)
            burnished_division = left_part + middle_part + right_part
            burnished_divisions.append(burnished_division)
        else:
            # first division
            available_left_count = len(divisions[0])
            left_count = min([left_count, available_left_count])
            middle_count = len(divisions[0]) - left_count
            left = left[:left_count]
            if not middle_classes:
                middle_classes = [1]
            middle = [middle_classes[0]]
            middle = middle_count * middle
            left_part, middle_part = sequencetools.Sequence(
                    divisions[0]).partition_by_counts(
                    [left_count, middle_count],
                    cyclic=False,
                    overhang=Exact,
                    )
            left_part = class_._burnish_division_part(left_part, left)
            middle_part = class_._burnish_division_part(middle_part, middle)
            burnished_division = left_part + middle_part
            burnished_divisions.append(burnished_division)
            # middle divisions
            for division in divisions[1:-1]:
                middle_part = division
                middle = len(division) * [middle_classes[0]]
                middle_part = class_._burnish_division_part(
                    middle_part,
                    middle,
                    )
                burnished_division = middle_part
                burnished_divisions.append(burnished_division)
            # last division:
            available_right_count = len(divisions[-1])
            right_count = min([right_count, available_right_count])
            middle_count = len(divisions[-1]) - right_count
            right = right[:right_count]
            middle = middle_count * [middle_classes[0]]
            middle_part, right_part = sequencetools.Sequence(
                    divisions[-1]).partition_by_counts(
                    [middle_count, right_count],
                    cyclic=False,
                    overhang=Exact,
                    )
            middle_part = class_._burnish_division_part(middle_part, middle)
            right_part = class_._burnish_division_part(right_part, right)
            burnished_division = middle_part + right_part
            burnished_divisions.append(burnished_division)
        unburnished_weights = [mathtools.weight(x) for x in divisions]
        burnished_weights = [mathtools.weight(x) for x in burnished_divisions]
        #assert burnished_weights == unburnished_weights
        # TODO: make the following work on Python 3:
        #assert tuple(burnished_weights) == tuple(unburnished_weights)
        return burnished_divisions

    def _get_format_specification(self):
        from abjad.tools import systemtools
        agent = systemtools.StorageFormatAgent(self)
        names = list(agent.signature_keyword_names)
        for name in names[:]:
            if not getattr(self, name):
                names.remove(name)
        return systemtools.FormatSpecification(
            client=self,
            storage_format_kwargs_names=names,
            )

    @staticmethod
    def _is_length_tuple(argument):
        if argument is None:
            return True
        if mathtools.all_are_nonnegative_integer_equivalent_numbers(argument):
            if isinstance(argument, tuple):
                return True
        return False

    @staticmethod
    def _is_sign_tuple(argument):
        from abjad.tools import scoretools
        if argument is None:
            return True
        if isinstance(argument, tuple):
            prototype = (-1, 0, 1, scoretools.Note, scoretools.Rest)
            return all(_ in prototype for _ in argument)
        return False

    def _none_to_trivial_helper(self, argument):
        if argument is None:
            argument = self._trivial_helper
        assert callable(argument)
        return argument

    def _rotate_input(self, helper_functions=None, rotation=None):
        helper_functions = helper_functions or {}
        input_ = {}
        names = (
            'left_classes',
            'left_counts',
            'middle_classes',
            'right_classes',
            'right_counts',
            )
        for name in names:
            value = getattr(self, name)
            value = value or ()
            helper_function = helper_functions.get(name)
            helper_function = self._none_to_trivial_helper(helper_function)
            value = helper_function(value, rotation)
            value = datastructuretools.CyclicTuple(value)
            input_[name] = value
        return input_

    def _trivial_helper(self, sequence_, rotation):
        if isinstance(rotation, int) and len(sequence_):
            return sequencetools.Sequence(sequence_).rotate(n=rotation)
        return sequence_

    ### PUBLIC PROPERTIES ###

    @property
    def left_classes(self):
        r'''Gets left classes.

        ..  container:: example

            ::

                >>> burnish_specifier = rhythmmakertools.BurnishSpecifier(
                ...     left_classes=[Rest, 0],
                ...     right_classes=[Rest, Rest, 0],
                ...     left_counts=[2],
                ...     right_counts=[1],
                ...     )

            ::

                >>> burnish_specifier.left_classes
                [<class 'abjad.tools.scoretools.Rest.Rest'>, 0]

        Returns tuple or none.
        '''
        if self._left_classes:
            return list(self._left_classes)

    @property
    def left_counts(self):
        r'''Gets left counts.

        ..  container:: example

            ::

                >>> burnish_specifier = rhythmmakertools.BurnishSpecifier(
                ...     left_classes=[Rest, 0],
                ...     right_classes=[Rest, Rest, 0],
                ...     left_counts=[2],
                ...     right_counts=[1],
                ...     )

            ::

                >>> burnish_specifier.left_counts
                [2]

        Returns tuple or none.
        '''
        if self._left_counts:
            return list(self._left_counts)

    @property
    def middle_classes(self):
        r'''Gets middle_classes.

        ..  container:: example

            ::

                >>> burnish_specifier = rhythmmakertools.BurnishSpecifier(
                ...     left_classes=[Rest, 0],
                ...     right_classes=[Rest, Rest, 0],
                ...     left_counts=[2],
                ...     right_counts=[1],
                ...     )

            ::

                >>> burnish_specifier.middle_classes is None
                True

        Returns tuple or none.
        '''
        if self._middle_classes:
            return list(self._middle_classes)

    @property
    def outer_divisions_only(self):
        r'''Is true when rhythm-maker should burnish only first and last 
        division in output.
        
        Is false when rhythm-maker should burnish all divisions.

        Defaults to false.

        Set to true or false.
        '''
        return self._outer_divisions_only

    @property
    def right_classes(self):
        r'''Gets right classes.

        ..  container:: example

            ::

                >>> burnish_specifier = rhythmmakertools.BurnishSpecifier(
                ...     left_classes=[Rest, 0],
                ...     right_classes=[Rest, Rest, 0],
                ...     left_counts=[2],
                ...     right_counts=[1],
                ...     )

            ::

                >>> burnish_specifier.right_classes
                [<class 'abjad.tools.scoretools.Rest.Rest'>, <class 'abjad.tools.scoretools.Rest.Rest'>, 0]

        Returns tuple or none.
        '''
        if self._right_classes:
            return list(self._right_classes)

    @property
    def right_counts(self):
        r'''Gets right counts.

        ..  container:: example

            ::

                >>> burnish_specifier = rhythmmakertools.BurnishSpecifier(
                ...     left_classes=[Rest, 0],
                ...     right_classes=[Rest, Rest, 0],
                ...     left_counts=[2],
                ...     right_counts=[1],
                ...     )

            ::

                >>> burnish_specifier.right_counts
                [1]

        Returns tuple or none.
        '''
        if self._right_counts:
            return list(self._right_counts)



# -*- coding: utf-8 -*-
import platform
import pytest
import sys
from abjad.tools import mathtools
from abjad.tools import systemtools


def test_mathtools_NonreducedFraction___add___01():

    one = mathtools.NonreducedFraction(1, 4)
    two = mathtools.NonreducedFraction(2, 8)
    result = one + two
    assert result.pair == (4, 8)

    one = mathtools.NonreducedFraction(2, 8)
    two = mathtools.NonreducedFraction(1, 4)
    result = one + two
    assert result.pair == (4, 8)


def test_mathtools_NonreducedFraction___add___02():

    result = mathtools.NonreducedFraction(3, 3) + 1
    assert result.pair == (6, 3)

    result = 1 + mathtools.NonreducedFraction(3, 3)
    assert result.pair == (6, 3)


@pytest.mark.skipif(
    platform.python_implementation() != 'CPython',
    reason='Benchmarking is only for CPython.',
    )
def test_mathtools_NonreducedFraction___add___03():

    a = mathtools.NonreducedFraction(3, 6)
    b = mathtools.NonreducedFraction(3, 12)

    result_one = systemtools.IOManager.count_function_calls(
        'a + b', locals())
    result_two = systemtools.IOManager.count_function_calls(
        'a + 10', locals())

    if sys.version_info[0] == 2:
        assert result_one <= 80
        assert result_two <= 80
    else:
        assert result_one <= 100
        assert result_two <= 100



# -*- coding: utf-8 -*-
from abjad.tools.abctools import AbjadObject


class LilyPondFormatBundle(AbjadObject):
    r'''LilyPond format bundle.

    Transient class created to hold the collection of all
    format contributions generated on behalf of a single component.
    '''

    ### CLASS VARIABLES ###

    __documentation_section__ = 'LilyPond formatting'

    __slots__ = (
        '_before',
        '_after',
        '_opening',
        '_closing',
        '_right',
        '_context_settings',
        '_grob_overrides',
        '_grob_reverts',
        )

    # TODO: implement as full-fledged class in its own right in systemtools
    ### INNER CLASS DEFINITION ###

    class SlotContributions(AbjadObject):
        r'''Slot contributions.
        '''

        __slots__ = (
            '_articulations',
            '_commands',
            '_comments',
            '_indicators',
            '_markup',
            '_spanners',
            '_spanner_starts',
            '_spanner_stops',
            '_stem_tremolos',
            '_trill_pitches',
            )

        def __init__(self):
            self._articulations = []
            self._commands = []
            self._comments = []
            self._indicators = []
            self._markup = []
            self._spanners = []
            self._spanner_starts = []
            self._spanner_stops = []
            self._stem_tremolos = []
            self._trill_pitches = []

        def _get_format_specification(self):
            from abjad.tools import systemtools
            names = [
                'articulations',
                'commands',
                'comments',
                'indicators',
                'markup',
                'spanners',
                'spanner_starts',
                'spanner_stops',
                'stem_tremolos',
                'trill_pitches',
                ]
            names = [x for x in names if getattr(self, x)]
            return systemtools.FormatSpecification(
                client=self,
                storage_format_kwargs_names=names,
                )

        @property
        def articulations(self):
            return self._articulations

        @property
        def comments(self):
            return self._comments

        @property
        def commands(self):
            return self._commands

        @property
        def has_contributions(self):
            contribution_categories = (
                'articulations',
                'commands',
                'comments',
                'indicators',
                'markup',
                'spanners',
                'spanner_starts',
                'spanner_stops',
                'stem_tremolos',
                'trill_pitches',
                )
            return any(getattr(self, contribution_category)
                for contribution_category in contribution_categories)

        @property
        def indicators(self):
            return self._indicators

        @property
        def markup(self):
            return self._markup

        @property
        def spanners(self):
            return self._spanners

        @property
        def spanner_starts(self):
            return self._spanner_starts

        @property
        def spanner_stops(self):
            return self._spanner_stops

        @property
        def stem_tremolos(self):
            return self._stem_tremolos

        @property
        def trill_pitches(self):
            return self._trill_pitches

        def alphabetize(self):
            self._indicators.sort()

        def get(self, identifier):
            return getattr(self, identifier)

        def make_immutable(self):
            self._articulations = tuple(sorted(self.articulations))
            self._commands = tuple(self.commands)
            self._comments = tuple(self.comments)
            self._indicators = tuple(self.indicators)
            self._markup = tuple(self.markup)
            self._spanners = tuple(self.spanners)
            self._spanner_starts = tuple(self.spanner_starts)
            self._spanner_stops = tuple(self.spanner_stops)
            self._stem_tremolos = tuple(self.stem_tremolos)
            self._trill_pitches = tuple(self.trill_pitches)

        def update(self, slot_contributions):
            assert isinstance(slot_contributions, type(self))
            self.articulations.extend(slot_contributions.articulations)
            self.commands.extend(slot_contributions.commands)
            self.comments.extend(slot_contributions.comments)
            self.indicators.extend(slot_contributions.indicators)
            self.markup.extend(slot_contributions.markup)
            self.spanners.extend(slot_contributions.spanners)
            self.spanner_starts.extend(slot_contributions.spanner_starts)
            self.spanner_stops.extend(slot_contributions.spanner_stops)
            self.stem_tremolos.extend(slot_contributions.stem_tremolos)
            self.trill_pitches.extend(slot_contributions.trill_pitches)

    ### INITIALIZER ###

    def __init__(self):
        self._before = self.SlotContributions()
        self._after = self.SlotContributions()
        self._opening = self.SlotContributions()
        self._closing = self.SlotContributions()
        self._right = self.SlotContributions()
        self._context_settings = []
        self._grob_overrides = []
        self._grob_reverts = []

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        from abjad.tools import systemtools
        slot_contribution_names = (
            'before',
            'after',
            'opening',
            'closing',
            'right',
            )
        grob_contribution_names = (
            'context_settings',
            'grob_overrides',
            'grob_reverts',
            )
        names = [x for x in slot_contribution_names
            if getattr(self, x).has_contributions]
        names.extend(x for x in grob_contribution_names
            if getattr(self, x))
        return systemtools.FormatSpecification(
            client=self,
            storage_format_kwargs_names=names,
            )

    ### PUBLIC METHODS ###

    def alphabetize(self):
        r'''Alphabetizes format contributions in each slot.

        Returns none.
        '''
        self.before.alphabetize()
        self.after.alphabetize()
        self.opening.alphabetize()
        self.closing.alphabetize()
        self.right.alphabetize()
        self._context_settings.sort()
        self._grob_overrides.sort()
        self._grob_reverts.sort()

    def get(self, identifier):
        r'''Gets `identifier`.

        Returns format contributions object or list.
        '''
        return getattr(self, identifier)

    def make_immutable(self):
        r'''Makes each slot immutable.

        Returns none.
        '''
        self.before.make_immutable()
        self.after.make_immutable()
        self.opening.make_immutable()
        self.closing.make_immutable()
        self.right.make_immutable()
        self._context_settings = tuple(sorted(set(self.context_settings)))
        self._grob_overrides = tuple(sorted(set(self.grob_overrides)))
        self._grob_reverts = tuple(sorted(set(self.grob_reverts)))

    def update(self, format_bundle):
        r'''Updates format bundle with all format contributions in
        `format_bundle`.

        Returns none.
        '''
        if hasattr(format_bundle, '_get_lilypond_format_bundle'):
            format_bundle = format_bundle._get_lilypond_format_bundle()
        assert isinstance(format_bundle, type(self))
        self.before.update(format_bundle.before)
        self.after.update(format_bundle.after)
        self.opening.update(format_bundle.opening)
        self.closing.update(format_bundle.closing)
        self.right.update(format_bundle.right)
        self.context_settings.extend(format_bundle.context_settings)
        self.grob_overrides.extend(format_bundle.grob_overrides)
        self.grob_reverts.extend(format_bundle.grob_reverts)

    ### PUBLIC PROPERTIES ###

    @property
    def after(self):
        r'''After slot contributions.

        Returns slot contributions object.
        '''
        return self._after

    @property
    def before(self):
        r'''Before slot contributions.

        Returns slot contributions object.
        '''
        return self._before

    @property
    def closing(self):
        r'''Closing slot contributions.

        Returns slot contributions object.
        '''
        return self._closing

    @property
    def context_settings(self):
        r'''Context setting format contributions.

        Returns list.
        '''
        return self._context_settings

    @property
    def grob_overrides(self):
        r'''Grob override format contributions.

        Returns list.
        '''
        return self._grob_overrides

    @property
    def grob_reverts(self):
        r'''Grob revert format contributions.

        Returns list.
        '''
        return self._grob_reverts

    @property
    def opening(self):
        r'''Opening slot contributions.

        Returns slot contributions object.
        '''
        return self._opening

    @property
    def right(self):
        r'''Right slot contributions.

        Returns slot contributions object.
        '''
        return self._right



# -*- coding: utf-8 -*-
from abjad import *


def test_pitchtools_NamedPitchClass_transpose_01():

    npc = pitchtools.NamedPitchClass('c')

    new = npc.transpose(pitchtools.NamedInterval('perfect', 1))
    new == pitchtools.NamedPitchClass('c')

    new = npc.transpose(pitchtools.NamedInterval('minor', 2))
    new == pitchtools.NamedPitchClass('df')

    new = npc.transpose(pitchtools.NamedInterval('minor', -2))
    new == pitchtools.NamedPitchClass('b')

    new = npc.transpose(pitchtools.NamedInterval('major', 2))
    new == pitchtools.NamedPitchClass('d')

    new = npc.transpose(pitchtools.NamedInterval('major', -2))
    new == pitchtools.NamedPitchClass('bf')



# -*- coding: utf-8 -*-
from abjad import *
import pytest


def test_scoretools_Container_insert_01():
    r'''Insert component into container at index i.
    Fracture spanners to the left of index i.
    Fracture spanners to the right of index i.
    Returns Python list of fractured spanners.
    '''

    "Insert works just before a spanner."

    voice = Voice("c'8 d'8 e'8 f'8")
    beam = Beam()
    attach(beam, voice[:])
    voice.insert(0, Rest((1, 8)))

    r'''
    \new Voice {
        r8
        c'8 [
        d'8
        e'8
        f'8 ]
    }
    '''

    assert inspect_(voice).is_well_formed()
    assert format(voice) == stringtools.normalize(
        r'''
        \new Voice {
            r8
            c'8 [
            d'8
            e'8
            f'8 ]
        }
        '''
        )


def test_scoretools_Container_insert_02():
    r'''Insert works inside a spanner.
    '''

    voice = Voice("c'8 d'8 e'8 f'8")
    beam = Beam()
    attach(beam, voice[:])
    voice.insert(1, Note(1, (1, 8)))

    r'''
    \new Voice {
        c'8 [
        cs'8
        d'8
        e'8
        f'8 ]
    }
    '''

    assert inspect_(voice).is_well_formed()
    assert format(voice) == stringtools.normalize(
        r'''
        \new Voice {
            c'8 [
            cs'8
            d'8
            e'8
            f'8 ]
        }
        '''
        )



def test_scoretools_Container_insert_03():
    r'''Insert works just after a spanner.
    '''

    staff = Staff([Note(n, (1, 8)) for n in range(4)])
    beam = Beam()
    attach(beam, staff[:])
    staff.insert(4, Rest((1, 4)))

    r'''
    \new Staff {
        c'8 [
        cs'8
        d'8
        ef'8 ]
        r4
    }
    '''

    assert inspect_(staff).is_well_formed()
    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff {
            c'8 [
            cs'8
            d'8
            ef'8 ]
            r4
        }
        '''
        )


def test_scoretools_Container_insert_04():
    r'''Insert works with really big positive values.
    '''

    staff = Staff([Note(n, (1, 8)) for n in range(4)])
    beam = Beam()
    attach(beam, staff[:])
    staff.insert(1000, Rest((1, 4)))

    r'''
    \new Staff {
        c'8 [
        cs'8
        d'8
        ef'8 ]
    r4
    }
    '''

    assert inspect_(staff).is_well_formed()
    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff {
            c'8 [
            cs'8
            d'8
            ef'8 ]
            r4
        }
        '''
        )


def test_scoretools_Container_insert_05():
    r'''Insert works with negative values.
    '''

    voice = Voice("c'8 d'8 e'8 f'8")
    beam = Beam()
    attach(beam, voice[:])
    voice.insert(-1, Note(4.5, (1, 8)))

    r'''
    \new Voice {
        c'8 [
        d'8
        e'8
        eqs'8
        f'8 ]
    }
    '''

    assert inspect_(voice).is_well_formed()
    assert format(voice) == stringtools.normalize(
        r'''
        \new Voice {
            c'8 [
            d'8
            e'8
            eqs'8
            f'8 ]
        }
        '''
        )


def test_scoretools_Container_insert_06():
    r'''Insert works with really big negative values.
    '''

    voice = Voice("c'8 d'8 e'8 f'8")
    beam = Beam()
    attach(beam, voice[:])
    voice.insert(-1000, Rest((1, 8)))

    r'''
    \new Voice {
        r8
        c'8 [
        d'8
        e'8
        f'8 ]
    }
    '''

    assert inspect_(voice).is_well_formed()
    assert format(voice) == stringtools.normalize(
        r'''
        \new Voice {
            r8
            c'8 [
            d'8
            e'8
            f'8 ]
        }
        '''
        )


def test_scoretools_Container_insert_07():
    r'''Inserting a note from one container into another container
    changes note parent from first container to second.
    '''

    voice = Voice("c'8 d'8 e'8 f'8")
    staff = Staff("c'8 c'8 c'8 c'8 c'8 c'8 c'8 c'8")
    note = voice[0]
    staff.insert(1, voice[0])

    assert inspect_(voice).is_well_formed()
    assert inspect_(staff).is_well_formed()
    assert not note in voice
    assert note._parent is staff


def test_scoretools_Container_insert_08():

    staff = Staff("c'8 d'8 e'8 f'8")
    beam = Beam()
    attach(beam, staff[:])
    staff.insert(1, Note("cs'8"), fracture_spanners=False)

    r'''
    \new Staff {
        c'8 [
        cs'8
        d'8
        e'8
        f'8 ]
    }
    '''

    assert inspect_(staff).is_well_formed()
    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff {
            c'8 [
            cs'8
            d'8
            e'8
            f'8 ]
        }
        '''
        )


def test_scoretools_Container_insert_09():
    r'''Insert component into container at index i.
    Fracture spanners to the left of index i.
    Fracture spanners to the right of index i.
    Returns Python list of fractured spanners.
    '''

    "Insert works just before a spanner."

    staff = Staff([Note(n, (1, 8)) for n in range(4)])
    beam = Beam()
    attach(beam, staff[:])
    staff.insert(0, Rest('r4'), fracture_spanners=True)

    r'''
    \new Staff {
        r4
        c'8 [
        cs'8
        d'8
        ef'8 ]
    }
    '''

    assert inspect_(staff).is_well_formed()
    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff {
            r4
            c'8 [
            cs'8
            d'8
            ef'8 ]
        }
        '''
        )


def test_scoretools_Container_insert_10():
    r'''Insert works inside a spanner.
    '''

    staff = Staff([Note(n, (1, 8)) for n in range(4)])
    beam = Beam()
    attach(beam, staff[:])
    staff.insert(1, Rest('r4'), fracture_spanners=True)

    r'''
    \new Staff {
        c'8 [ ]
        r4
        cs'8 [
        d'8
        ef'8 ]
    }
    '''

    assert inspect_(staff).is_well_formed()
    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff {
            c'8 [ ]
            r4
            cs'8 [
            d'8
            ef'8 ]
        }
        '''
        )


def test_scoretools_Container_insert_11():
    r'''Insert works just after a spanner.
    '''

    staff = Staff([Note(n, (1, 8)) for n in range(4)])
    beam = Beam()
    attach(beam, staff[:])
    staff.insert(4, Rest('r4'), fracture_spanners=True)

    r'''
    \new Staff {
        c'8 [
        cs'8
        d'8
        ef'8 ]
        r4
    }
    '''

    assert inspect_(staff).is_well_formed()
    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff {
            c'8 [
            cs'8
            d'8
            ef'8 ]
            r4
        }
        '''
        )


def test_scoretools_Container_insert_12():
    r'''Insert works with really big positive values.
    '''

    staff = Staff([Note(n, (1, 8)) for n in range(4)])
    beam = Beam()
    attach(beam, staff[:])
    staff.insert(1000, Rest('r4'), fracture_spanners=True)

    r'''
    \new Staff {
        c'8 [
        cs'8
        d'8
        ef'8 ]
    r4
    }
    '''

    assert inspect_(staff).is_well_formed()
    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff {
            c'8 [
            cs'8
            d'8
            ef'8 ]
            r4
        }
        '''
        )


def test_scoretools_Container_insert_13():
    r'''Insert works with negative values.
    '''

    staff = Staff([Note(n, (1, 8)) for n in range(4)])
    beam = Beam()
    attach(beam, staff[:])
    staff.insert(-1, Rest('r4'), fracture_spanners=True)

    r'''
    \new Staff {
        c'8 [
        cs'8
        d'8 ]
        r4
        ef'8 [ ]
    }
    '''

    assert inspect_(staff).is_well_formed()
    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff {
            c'8 [
            cs'8
            d'8 ]
            r4
            ef'8 [ ]
        }
        '''
        )


def test_scoretools_Container_insert_14():
    r'''Insert works with really big negative values.
    '''

    staff = Staff([Note(n, (1, 8)) for n in range(4)])
    beam = Beam()
    attach(beam, staff[:])
    staff.insert(-1000, Rest('r4'), fracture_spanners=True)

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff {
            r4
            c'8 [
            cs'8
            d'8
            ef'8 ]
        }
        '''
        )

    assert inspect_(staff).is_well_formed()


def test_scoretools_Container_insert_15():
    r'''Inserting a note from one container into another container
    changes note parent from first container to second.
    '''

    voice = Voice("c'8 d'8 e'8 f'8")
    staff = Staff("c'8 c'8 c'8 c'8 c'8 c'8 c'8 c'8")
    note = voice[0]
    staff.insert(1, voice[0], fracture_spanners=True)

    assert inspect_(voice).is_well_formed()
    assert inspect_(staff).is_well_formed()
    assert not note in voice
    assert note._parent is staff


# -*- coding: utf-8 -*-
import abc
import copy
from abjad.tools import sequencetools
from abjad.tools.abctools import AbjadObject


class SearchTree(AbjadObject):
    r'''Abstract base class from which concrete ``SearchTree`` subclasses
    inherit.

    ``SearchTrees`` encapsulate strategies for generating collections of
    ``QGrids``, given a set of ``QEventProxy`` instances as input.

    They allow composers to define the degree and quality of nested rhythmic
    subdivisions in the quantization output.  That is to say, they allow
    composers to specify what sorts of tuplets and ratios of pulses may be
    contained within other tuplets, to arbitrary levels of nesting.
    '''

    ### CLASS VARIABLES ###

    __slots__ = (
        '_definition',
        )

    ### INITIALIZER ###

    def __init__(self, definition=None):
        if definition is None:
            definition = self.default_definition
        else:
            assert self._is_valid_definition(definition)
        self._definition = definition

    ### SPECIAL METHODS ###

    def __call__(self, q_grid):
        r'''Calls search tree.
        '''
        from abjad.tools import quantizationtools
        assert isinstance(q_grid, quantizationtools.QGrid)
        new_q_grids = []
        commands = self._generate_all_subdivision_commands(q_grid)
        for command in commands:
            new_q_grid = copy.copy(q_grid)
            q_events = new_q_grid.subdivide_leaves(command)
            new_q_grid.fit_q_events(q_events)
            new_q_grids.append(new_q_grid)
        return new_q_grids

    def __eq__(self, argument):
        r'''Is true when `argument` is a search tree with definition equal to that of
        this search tree. Otherwise false.

        Returns true or false.
        '''
        if type(self) == type(argument):
            if self.definition == argument.definition:
                return True
        return False

    def __hash__(self):
        r'''Hashes search tree.

        Required to be explicitly redefined on Python 3 if __eq__ changes.

        Returns integer.
        '''
        return super(SearchTree, self).__hash__()

    ### PRIVATE METHODS ###

    def _find_divisible_leaf_indices_and_subdivisions(self, q_grid):
        # TODO: This should actually check for all QEvents which fall
        # within the leaf's duration,
        # including QEvents attached to the next leaf
        # It may be prudent to actually store QEvents in two lists:
        # before_offset and after_offset
        indices, subdivisions = [], []
        leaves = list(q_grid.leaves)
        i = 0
        for leaf_one, leaf_two in sequencetools.Sequence(leaves).nwise():
            if leaf_one.is_divisible:
                succeeding_proxies = leaf_one.succeeding_q_event_proxies
                preceding_proxies = leaf_two.preceding_q_event_proxies
                if not preceding_proxies and \
                    all(proxy.offset == leaf_one.start_offset
                        for proxy in succeeding_proxies):
                    pass  # proxies align perfectly with this leaf

                elif preceding_proxies or succeeding_proxies:
                    parentage_ratios = leaf_one.parentage_ratios
                    leaf_subdivisions = \
                        self._find_leaf_subdivisions(parentage_ratios)
                    if leaf_subdivisions:
                        indices.append(i)
                        subdivisions.append(tuple(leaf_subdivisions))
            i += 1
        return indices, subdivisions

    @abc.abstractmethod
    def _find_leaf_subdivisions(self, leaf):
        raise NotImplementedError

    def _generate_all_subdivision_commands(self, q_grid):
        indices, subdivisions = \
            self._find_divisible_leaf_indices_and_subdivisions(q_grid)
        if not indices:
            return ()
        enumeration = sequencetools.Enumeration(subdivisions)
        combinations = enumeration.yield_outer_product()
        combinations = [tuple(_) for _ in combinations]
        return tuple(tuple(zip(indices, combo)) for combo in combinations)

    @abc.abstractmethod
    def _is_valid_definition(self, definition):
        raise NotImplementedError

    ### PUBLIC PROPERTIES ###

    @abc.abstractproperty
    def default_definition(self):
        r'''The default search tree definition.

        Returns dictionary.
        '''
        raise NotImplementedError

    @property
    def definition(self):
        r'''The search tree definition.

        Returns dictionary.
        '''
        return self._definition



# -*- coding: utf-8 -*-
import pytest
from abjad import *
from abjad.tools.lilypondparsertools import LilyPondParser


def test_lilypondparsertools_LilyPondParser__spanners__PhrasingSlur_01():
    r'''Successful slurs, showing single leaf overlap.
    '''

    target = Container(scoretools.make_notes([0] * 4, [(1, 4)]))
    slur = spannertools.PhrasingSlur()
    attach(slur, target[2:])
    slur = spannertools.PhrasingSlur()
    attach(slur, target[:3])

    assert format(target) == stringtools.normalize(
        r'''
        {
            c'4 \(
            c'4
            c'4 \) \(
            c'4 \)
        }
        '''
        )

    parser = LilyPondParser()
    result = parser(format(target))
    assert format(target) == format(result) and target is not result


def test_lilypondparsertools_LilyPondParser__spanners__PhrasingSlur_02():
    r'''Swapped start and stop.
    '''

    target = Container(scoretools.make_notes([0] * 4, [(1, 4)]))
    slur = spannertools.PhrasingSlur()
    attach(slur, target[2:])
    slur = spannertools.PhrasingSlur()
    attach(slur, target[:3])

    assert format(target) == stringtools.normalize(
        r'''
        {
            c'4 \(
            c'4
            c'4 \) \(
            c'4 \)
        }
        '''
        )

    string = r"\relative c' { c \( c c \( \) c \) }"

    parser = LilyPondParser()
    result = parser(string)
    assert format(target) == format(result) and target is not result


def test_lilypondparsertools_LilyPondParser__spanners__PhrasingSlur_03():
    r'''Single leaf.
    '''

    string = '{ c \( \) c c c }'
    assert pytest.raises(Exception, 'LilyPondParser()(string)')


def test_lilypondparsertools_LilyPondParser__spanners__PhrasingSlur_04():
    r'''Unterminated.
    '''

    string = '{ c \( c c c }'
    assert pytest.raises(Exception, 'LilyPondParser()(string)')


def test_lilypondparsertools_LilyPondParser__spanners__PhrasingSlur_05():
    r'''Unstarted.
    '''

    string = '{ c c c c \) }'
    assert pytest.raises(Exception, 'LilyPondParser()(string)')


def test_lilypondparsertools_LilyPondParser__spanners__PhrasingSlur_06():
    r'''Nested.
    '''

    string = '{ c \( c \( c \) c \) }'
    assert pytest.raises(Exception, 'LilyPondParser()(string)')



# -*- coding: utf-8 -*-
from abjad import *


def test_scoretools_FixedDurationTuplet_is_trivial_01():
    r'''Is true when tuplet ratio equals one.
    '''

    tuplet = scoretools.FixedDurationTuplet(Duration(2, 8), "c'8 d'8")
    assert tuplet.is_trivial


def test_scoretools_FixedDurationTuplet_is_trivial_02():
    r'''False when tuplet ratio does not equal one.
    '''

    tuplet = scoretools.FixedDurationTuplet(Duration(2, 8), "c'8 d'8 e'8")
    assert not tuplet.is_trivial



# -*- coding: utf-8 -*-
from abjad import *


def test_scoretools_Measure_duration_is_underfull_01():

    measure = Measure((3, 8), "c'8 c'8 c'8")
    assert not measure.is_underfull

    detach(TimeSignature, measure)
    time_signature = TimeSignature((4, 8))
    attach(time_signature, measure)
    assert measure.is_underfull

    detach(TimeSignature, measure)
    time_signature = TimeSignature((3, 8))
    attach(time_signature, measure)
    assert not measure.is_underfull



# -*- coding: utf-8 -*-
from abjad import *


def test_instrumenttools_iterate_out_of_range_notes_and_chords_01():

    staff = Staff("c'8 r8 <d fs>8 r8")
    violin = instrumenttools.Violin()
    attach(violin, staff)

    leaves = instrumenttools.iterate_out_of_range_notes_and_chords(staff)
    leaves = list(leaves)

    assert len(leaves) == 1
    assert leaves[0] is staff[2]



# -*- coding: utf-8 -*-
from abjad import *
from abjad.tools.lilypondparsertools import LilyPondParser


def test_lilypondparsertools_LilyPondParser__contexts__Staff_01():

    target = Staff([])

    assert format(target) == stringtools.normalize(
        r'''
        \new Staff {
        }
        '''
        )

    parser = LilyPondParser()
    result = parser(format(target))
    assert format(target) == format(result) and target is not result


def test_lilypondparsertools_LilyPondParser__contexts__Staff_02():

    target = Staff([])
    target.is_simultaneous = True
    target.append(Voice(scoretools.make_notes([0, 2, 4, 5, 7, 9, 11, 12], (1, 8))))
    target.append(Voice(scoretools.make_notes([0, 2, 4, 5, 7, 9, 11, 12], (1, 8))))

    assert format(target) == stringtools.normalize(
        r'''
        \new Staff <<
            \new Voice {
                c'8
                d'8
                e'8
                f'8
                g'8
                a'8
                b'8
                c''8
            }
            \new Voice {
                c'8
                d'8
                e'8
                f'8
                g'8
                a'8
                b'8
                c''8
            }
        >>
        '''
        )

    parser = LilyPondParser()
    result = parser(format(target))
    assert format(target) == format(result) and target is not result



# -*- coding: utf-8 -*-
import inspect
import pytest
from abjad.tools import abjadbooktools
from abjad.tools import documentationtools
from abjad.tools import expressiontools
from abjad.tools import systemtools


ignored_classes = (
    abjadbooktools.CodeBlock,
    abjadbooktools.CodeOutputProxy,
    abjadbooktools.GraphvizOutputProxy,
    abjadbooktools.LilyPondOutputProxy,
    expressiontools.Expression,
    systemtools.StorageFormatAgent,
    systemtools.FormatSpecification,
    )

classes = documentationtools.list_all_abjad_classes(
    ignored_classes=ignored_classes,
    )


@pytest.mark.parametrize('class_', classes)
def test_abjad___slots___01(class_):
    r'''All concrete classes defining __slots__ do not have a dict.
    '''
    if inspect.isabstract(class_):
        return
    attrs = inspect.classify_class_attrs(class_)
    attrs = dict((attr.name, attr) for attr in attrs)
    if '__slots__' not in attrs:
        return
    elif attrs['__slots__'].defining_class is not class_:
        return
    instance_one = class_()
    assert not hasattr(instance_one, '__dict__')



# -*- coding: utf-8 -*-
import functools
import inspect
import pytest
import sys
from abjad.tools import abjadbooktools
from abjad.tools import datastructuretools
from abjad.tools import documentationtools
from abjad.tools import durationtools
from abjad.tools import mathtools
from abjad.tools import systemtools


ignored_classes = (
    abjadbooktools.AbjadDirective,
    abjadbooktools.CodeBlock,
    abjadbooktools.CodeOutputProxy,
    abjadbooktools.DoctestDirective,
    abjadbooktools.GraphvizOutputProxy,
    abjadbooktools.ImportDirective,
    abjadbooktools.LilyPondOutputProxy,
    abjadbooktools.RevealDirective,
    abjadbooktools.ShellDirective,
    abjadbooktools.ThumbnailDirective,
    datastructuretools.Enumeration,
    systemtools.StorageFormatAgent,
    systemtools.FormatSpecification,
    systemtools.TestCase,
    )

classes = documentationtools.list_all_abjad_classes(
    ignored_classes=ignored_classes,
    )


@pytest.mark.parametrize('class_', classes)
def test_abjad___init___01(class_):
    r'''All concrete classes initialize from empty input.
    '''
    if inspect.isabstract(class_):
        return
    instance = class_()
    assert instance is not None


valid_types = (
    bool,
    float,
    int,
    str,
    tuple,
    type(None),
    datastructuretools.OrdinalConstant,
    durationtools.Duration,
    mathtools.Infinity,
    mathtools.NegativeInfinity,
    )


@pytest.mark.parametrize('obj', classes)
def test_abjad___init___02(obj):
    r'''Make sure class initializer keyword argument values are immutable.
    '''
    version = sys.version
    # NOTE: something changed in 3.5's "inspect" module
    if not version.startswith('3.5'):
        for attr in inspect.classify_class_attrs(obj):
            if attr.defining_class is not obj:
                continue
            elif attr.kind != 'method':
                continue
            obj = attr.object
            if isinstance(obj, functools.partial):
                obj = obj.function
            argument_specification = inspect.getargspec(obj)
            keyword_argument_names = argument_specification.args[1:]
            keyword_argument_values = argument_specification.defaults
            if keyword_argument_values is None:
                continue
            for name, value in zip(
                keyword_argument_names, keyword_argument_values):
                assert isinstance(value, valid_types), (attr.name, name, value)
                if isinstance(value, tuple):
                    assert all(isinstance(x, valid_types) for x in value)


functions = documentationtools.list_all_abjad_functions()


@pytest.mark.parametrize('obj', functions)
def test_abjad___init___03(obj):
    r'''Make sure function keyword argument values are immutable.
    '''
    if isinstance(obj, functools.partial):
        obj = obj.function
    argument_specification = inspect.getargspec(obj)
    keyword_argument_names = argument_specification.args[1:]
    keyword_argument_values = argument_specification.defaults
    if keyword_argument_values is None:
        return
    for name, value in zip(
        keyword_argument_names, keyword_argument_values):
        assert isinstance(value, valid_types), (name, value)
        if isinstance(value, tuple):
            assert all(isinstance(x, valid_types) for x in value)



# -*- coding: utf-8 -*-
from abjad.tools.timespantools.TimeRelation import TimeRelation


class TimespanTimespanTimeRelation(TimeRelation):
    r'''A timespan vs. timespan time relation.

    Score for examples:

    ::

        >>> staff_1 = Staff(
        ...     r"\times 2/3 { c'4 d'4 e'4 } \times 2/3 { f'4 g'4 a'4 }")
        >>> staff_2 = Staff("c'2. d'4")
        >>> score = Score([staff_1, staff_2])

    ..  doctest::

        >>> f(score)
        \new Score <<
            \new Staff {
                \times 2/3 {
                    c'4
                    d'4
                    e'4
                }
                \times 2/3 {
                    f'4
                    g'4
                    a'4
                }
            }
            \new Staff {
                c'2.
                d'4
            }
        >>

    ::

        >>> last_tuplet = staff_1[-1]
        >>> long_note = staff_2[0]

    ::

        >>> show(score) # doctest: +SKIP

    ..  container:: example

       **Example 1:**

        ::

            >>> timespantools.timespan_2_happens_during_timespan_1(
            ...     timespan_1=last_tuplet,
            ...     timespan_2=long_note,
            ...     )
            False

    ..  container:: example

       **Example 2:**

        ::

            >>> timespantools.timespan_2_intersects_timespan_1(
            ...     timespan_1=last_tuplet,
            ...     timespan_2=long_note,
            ...     )
            True

    ..  container:: example

       **Example 3:**

        ::

            >>> timespantools.timespan_2_is_congruent_to_timespan_1(
            ...     timespan_1=last_tuplet,
            ...     timespan_2=long_note,
            ...     )
            False

    ..  container:: example

       **Example 4:**

        ::

            >>> timespantools.timespan_2_overlaps_all_of_timespan_1(
            ...     timespan_1=last_tuplet,
            ...     timespan_2=long_note,
            ...     )
            False

    ..  container:: example

       **Example 5:**

        ::

            >>> timespantools.timespan_2_overlaps_start_of_timespan_1(
            ...     timespan_1=last_tuplet,
            ...     timespan_2=long_note,
            ...     )
            True

    ..  container:: example

       **Example 6:**

        ::

            >>> timespantools.timespan_2_overlaps_stop_of_timespan_1(
            ...     timespan_1=last_tuplet,
            ...     timespan_2=long_note,
            ...     )
            False

    ..  container:: example

       **Example 7:**

        ::

            >>> timespantools.timespan_2_starts_after_timespan_1_starts(
            ...     timespan_1=last_tuplet,
            ...     timespan_2=long_note,
            ...     )
            False

    ..  container:: example

       **Example 8:**

        ::

            >>> timespantools.timespan_2_starts_after_timespan_1_stops(
            ...     timespan_1=last_tuplet,
            ...     timespan_2=long_note,
            ...     )
            False

    Timespan / timespan time relations are immutable.
    '''

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Time relations'

    __slots__ = (
        '_timespan_1',
        '_timespan_2',
        )

    ### INITIALIZER ###

    def __init__(self, inequality=None, timespan_1=None, timespan_2=None):
        TimeRelation.__init__(self, inequality=inequality)
        self._timespan_1 = timespan_1
        self._timespan_2 = timespan_2

    ### SPECIAL METHODS ###

    # TODO: hoist to TimeRelation
    def __call__(self, timespan_1=None, timespan_2=None):
        r'''Evaluate time relation.

        ..  container:: example

            **Example 1.** Evaluate time relation without substitution:

            ::

                >>> timespan_1 = timespantools.Timespan(5, 15)
                >>> timespan_2 = timespantools.Timespan(10, 20)

            ::

                >>> time_relation = timespantools.timespan_2_starts_during_timespan_1(
                ...     timespan_1=timespan_1,
                ...     timespan_2=timespan_2,
                ...     hold=True,
                ...     )

            ::

                >>> print(format(time_relation))
                timespantools.TimespanTimespanTimeRelation(
                    inequality=timespantools.CompoundInequality(
                        [
                            timespantools.Inequality('timespan_1.start_offset <= timespan_2.start_offset'),
                            timespantools.Inequality('timespan_2.start_offset < timespan_1.stop_offset'),
                            ],
                        logical_operator='and',
                        ),
                    timespan_1=timespantools.Timespan(
                        start_offset=durationtools.Offset(5, 1),
                        stop_offset=durationtools.Offset(15, 1),
                        ),
                    timespan_2=timespantools.Timespan(
                        start_offset=durationtools.Offset(10, 1),
                        stop_offset=durationtools.Offset(20, 1),
                        ),
                    )

            ::

                >>> time_relation()
                True

        ..  container:: example

            **Example 2.** Substitute `timespan_1` during evaluation:

            ::

                >>> new_timespan_1 = timespantools.Timespan(0, 10)

            ::

                >>> new_timespan_1
                Timespan(start_offset=Offset(0, 1), stop_offset=Offset(10, 1))

            ::

                >>> time_relation(timespan_1=new_timespan_1)
                False

        ..  container:: example

            **Example 3.** Substitute `timespan_2` during evaluation:

            ::

                >>> new_timespan_2 = timespantools.Timespan(2, 12)

            ::

                >>> new_timespan_2
                Timespan(start_offset=Offset(2, 1), stop_offset=Offset(12, 1))

            ::

                >>> time_relation(timespan_2=new_timespan_2)
                False

        ..  container:: example

            **Example 4.** Substitute both `timespan_1` and `timespan_2`
            during evaluation:

            ::

                >>> time_relation(
                ...     timespan_1=new_timespan_1,
                ...     timespan_2=new_timespan_2,
                ...     )
                True

        Raise value error if either `timespan_1` or `timespan_2` is none.

        Otherwise return boolean.
        '''
        from abjad.tools import timespantools
        timespan_1 = timespan_1 or self.timespan_1
        timespan_2 = timespan_2 or self.timespan_2
        if timespan_1 is None or timespan_2 is None:
            message = 'time relation is not fully loaded: {!r}.'
            raise ValueError(message.format(self))
        if not isinstance(timespan_1, timespantools.Timespan):
            timespan_1 = timespantools.Timespan()._get_timespan(timespan_1)
        if not isinstance(timespan_2, timespantools.Timespan):
            timespan_2 = timespantools.Timespan()._get_timespan(timespan_2)
        truth_value = self.inequality.evaluate(
            timespan_1.start_offset, timespan_1.stop_offset,
            timespan_2.start_offset, timespan_2.stop_offset)
        return truth_value

    def __eq__(self, argument):
        r'''Is true when `argument` equals time relation. Otherwise false:

        ::

            >>> timespan = timespantools.Timespan(0, 10)
            >>> time_relation_1 = \
            ...     timespantools.timespan_2_starts_during_timespan_1()
            >>> time_relation_2 = \
            ...     timespantools.timespan_2_starts_during_timespan_1(
            ...     timespan_1=timespan)

        ::

            >>> time_relation_1 == time_relation_1
            True
            >>> time_relation_1 == time_relation_2
            False
            >>> time_relation_2 == time_relation_2
            True

        Returns true or false.
        '''
        if isinstance(argument, type(self)):
            if self.inequality == argument.inequality:
                if self.timespan_1 == argument.timespan_1:
                    if self.timespan_2 == argument.timespan_2:
                        return True
        return False

    def __hash__(self):
        r'''Hashes time relation.

        Required to be explicitly redefined on Python 3 if __eq__ changes.

        Returns integer.
        '''
        return super(TimespanTimespanTimeRelation, self).__hash__()

    ### PUBLIC METHODS ###

    def get_counttime_components(self, counttime_components):
        r'''Get `counttime_components` that satisfy `time_relation`:

        ::

            >>> voice = Voice(
            ...     [Note(i % 36, Duration(1, 4)) for i in range(200)])
            >>> timespan_1 = timespantools.Timespan(20, 22)
            >>> time_relation = \
            ...     timespantools.timespan_2_starts_during_timespan_1(
            ...     timespan_1=timespan_1)

        ::

            >>> result = time_relation.get_counttime_components(voice[:])

        ::

            >>> for counttime_component in result:
            ...     counttime_component
            Note("af'4")
            Note("a'4")
            Note("bf'4")
            Note("b'4")
            Note("c''4")
            Note("cs''4")
            Note("d''4")
            Note("ef''4")

        ::

            >>> result.get_timespan()
            Timespan(start_offset=Offset(20, 1), stop_offset=Offset(22, 1))

        `counttime_components` must belong to a single voice.

        `counttime_components` must be time-contiguous.

        The call shown here takes 78355 function calls under r9686.

        Returns selection.
        '''
        from abjad.tools import selectiontools
        from abjad.tools import timespantools

        # check input
        assert isinstance(counttime_components, (
            list, selectiontools.Selection)), repr(counttime_components)
        assert self.timespan_1 is not None

        # iterate counttime components
        result = []
        for counttime_component in counttime_components:
            if self(timespan_2=counttime_component._get_timespan()):
                result.append(counttime_component)

        # return result
        return selectiontools.Selection(result)

    def get_offset_indices(
        self, timespan_2_start_offsets, timespan_2_stop_offsets):
        r'''Get offset indices that satisfy time relation:

        ::

            >>> staff = Staff("c'8 d'8 e'8 f'8 g'8 a'8 b'8 c''8")
            >>> start_offsets = [inspect_(note).get_timespan().start_offset for note in staff]
            >>> stop_offsets = [inspect_(note).get_timespan().stop_offset for note in staff]

        ..  container:: example

            **Example 1.** Notes equal to ``staff[0:2]`` start
            during timespan ``[0, 3/16)``:

            ::

                >>> timespan_1 = timespantools.Timespan(Offset(0), Offset(3, 16))
                >>> time_relation = \
                ...     timespantools.timespan_2_starts_during_timespan_1(
                ...     timespan_1=timespan_1)
                >>> time_relation.get_offset_indices(start_offsets, stop_offsets)
                (0, 2)

        ..  container:: example

            **Example 2.** Notes equal to ``staff[2:8]`` start after timespan
            ``[0, 3/16)`` stops:

            ::

                >>> timespan_1 = timespantools.Timespan(Offset(0), Offset(3, 16))
                >>> time_relation = \
                ...     timespantools.timespan_2_starts_after_timespan_1_stops(
                ...     timespan_1=timespan_1)
                >>> time_relation.get_offset_indices(start_offsets, stop_offsets)
                (2, 8)

        Returns nonnegative integer pair.
        '''
        from abjad.tools import timespantools
        from abjad.tools import timespantools

        result = self.inequality.get_offset_indices(
            self.timespan_1,
            timespan_2_start_offsets,
            timespan_2_stop_offsets,
            )

        if not result:
            return []
        elif len(result) == 1:
            timespan = result[0]
            start_index = int(timespan.start_offset)
            stop_index = int(timespan.stop_offset)
            return start_index, stop_index
        elif 0 < len(result):
            message = 'inequality evaluates to disjunct range: {!r}.'
            message = message.format(result)
            raise Exception(message)

    ### PUBLIC PROPERTIES ###

    @property
    def is_fully_loaded(self):
        r'''Is true when `timespan_1` and `timespan_2` are both not none.
        Otherwise false:

        ::

            >>> timespan_1 = timespantools.Timespan(0, 10)
            >>> timespan_2 = timespantools.Timespan(5, 15)
            >>> time_relation = \
            ...     timespantools.timespan_2_starts_during_timespan_1(
            ...     timespan_1=timespan_1,
            ...     timespan_2=timespan_2,
            ...     hold=True,
            ...     )

        ::

            >>> time_relation.is_fully_loaded
            True

        Returns true or false.
        '''
        return self.timespan_1 is not None and self.timespan_2 is not None

    @property
    def is_fully_unloaded(self):
        r'''Is true when `timespan_1` and `timespan_2` are both none.
        Otherwise false.

            >>> time_relation.is_fully_unloaded
            False

        Returns true or false.
        '''
        return self.timespan_1 is None and self.timespan_2 is None

    @property
    def timespan_1(self):
        r'''Time relation timespan ``1``:

        ::

            >>> time_relation.timespan_1
            Timespan(start_offset=Offset(0, 1), stop_offset=Offset(10, 1))

        Returns timespan.
        '''
        return self._timespan_1

    @property
    def timespan_2(self):
        r'''Time relation timespan ``2``:

        ::

            >>> time_relation.timespan_2
            Timespan(start_offset=Offset(5, 1), stop_offset=Offset(15, 1))

        Returns timespan.
        '''
        return self._timespan_2



# -*- coding: utf-8 -*-
from abjad import *


def test_durationtools_Duration_durations_to_nonreduced_fractions_01():

    durations = [Fraction(2, 4), 3, (5, 16)]

    result = Duration.durations_to_nonreduced_fractions(durations)

    pairs = [
        mathtools.NonreducedFraction(x)
        for x in [(8, 16), (48, 16), (5, 16)]
        ]

    assert result == pairs



# -*- coding: utf-8 -*-
import copy
from abjad import *


def test_scoretools_Tuplet___copy___01():

    tuplet_1 = Tuplet((2, 3), "c'8 d'8 e'8")
    override(tuplet_1).note_head.color = 'red'

    assert format(tuplet_1) == stringtools.normalize(
        r'''
        \override NoteHead.color = #red
        \times 2/3 {
            c'8
            d'8
            e'8
        }
        \revert NoteHead.color
        '''
        )

    tuplet_2 = copy.copy(tuplet_1)

    assert format(tuplet_2) == stringtools.normalize(
        r'''
        \override NoteHead.color = #red
        \times 2/3 {
        }
        \revert NoteHead.color
        '''
        )

    assert not len(tuplet_2)


# -*- coding: utf-8 -*-
from abjad import *


def test_rhythmtreetools_RhythmTreeNode_root_01():

    leaf = rhythmtreetools.RhythmTreeLeaf()
    subsubcontainer = rhythmtreetools.RhythmTreeContainer()
    subcontainer = rhythmtreetools.RhythmTreeContainer()
    container = rhythmtreetools.RhythmTreeContainer()

    container.append(subcontainer)
    subcontainer.append(subsubcontainer)
    subsubcontainer.append(leaf)

    assert leaf.root == container
    assert subsubcontainer.root == container
    assert subcontainer.root == container
    assert container.root == container



# -*- coding: utf-8 -*-
from abjad import *
from abjad.tools import tonalanalysistools


def test_tonalanalysistools_Mode_named_interval_segment_01():

    m2 = pitchtools.NamedInterval('minor', 2)
    M2 = pitchtools.NamedInterval('major', 2)

    mdig = tonalanalysistools.Mode('dorian').named_interval_segment
    intervals = [M2, m2, M2, M2, M2, m2, M2]
    assert mdig == pitchtools.IntervalSegment(intervals)

    mdig = tonalanalysistools.Mode('phrygian').named_interval_segment
    intervals = [m2, M2, M2, M2, m2, M2, M2]
    assert mdig == pitchtools.IntervalSegment(intervals)

    mdig = tonalanalysistools.Mode('lydian').named_interval_segment
    intervals = [M2, M2, M2, m2, M2, M2, m2]
    assert mdig == pitchtools.IntervalSegment(intervals)



# -*- coding: utf-8 -*-


def to_tridirectional_direction_string(argument):
    r'''Changes `argument` to tridirectional direction string.

    ..  container:: example

        ::

            >>> stringtools.to_tridirectional_direction_string('^')
            'up'

    ..  container:: example

        ::

            >>> stringtools.to_tridirectional_direction_string('-')
            'center'

    ..  container:: example

        ::

            >>> stringtools.to_tridirectional_direction_string('_')
            'down'

    ..  container:: example

        ::

            >>> stringtools.to_tridirectional_direction_string(1)
            'up'

    ..  container:: example

        ::

            >>> stringtools.to_tridirectional_direction_string(0)
            'center'

    ..  container:: example

        ::

            >>> stringtools.to_tridirectional_direction_string(-1)
            'down'

    ..  container:: example

        ::

            >>> stringtools.to_tridirectional_direction_string('default')
            'center'

    Returns none when `argument` is none.

    Returns string or none.
    '''
    lookup = {
        Up: 'up',
        '^': 'up',
        'up': 'up',
        1: 'up',
        Down: 'down',
        '_': 'down',
        'down': 'down',
        -1: 'down',
        Center: 'center',
        '-': 'center',
        0: 'center',
        'center': 'center',
        'default': 'center',
        'neutral': 'center',
        }
    if argument is None:
        return None
    elif argument in lookup:
        return lookup[argument]
    raise ValueError(argument)



# -*- coding: utf-8 -*-
from abjad import *
from abjad.tools import tonalanalysistools


def test_tonalanalysistools_Mode___init___01():
    r'''Initialize with mode name.
    '''

    mode = tonalanalysistools.Mode('dorian')
    assert mode.mode_name == 'dorian'


def test_tonalanalysistools_Mode___init___02():
    r'''Initialize with other mode instance.
    '''

    mode = tonalanalysistools.Mode('dorian')
    new = tonalanalysistools.Mode(mode)

    assert new.mode_name == 'dorian'



# -*- coding: utf-8 -*-
from abjad import *


def test_indicatortools_LilyPondComment_after_01():
    r'''Test context comments after.
    '''

    voice = Voice("c'8 d'8 e'8 f'8")
    beam = Beam()
    attach(beam, voice[:])
    override(beam).beam.thickness = 3
    string = 'Voice after comments here.'
    comment = indicatortools.LilyPondComment(string, 'after')
    attach(comment, voice)
    string = 'More voice after comments.'
    comment = indicatortools.LilyPondComment(string, 'after')
    attach(comment, voice)

    assert format(voice) == stringtools.normalize(
        r'''
        \new Voice {
            \override Beam.thickness = #3
            c'8 [
            d'8
            e'8
            f'8 ]
            \revert Beam.thickness
        }
        % Voice after comments here.
        % More voice after comments.
        '''
        )

    assert inspect_(voice).is_well_formed()


def test_indicatortools_LilyPondComment_after_02():
    r'''Leaf comments after.
    '''

    note = Note(0, (1, 8))
    override(note).beam.thickness = 3
    string = 'Leaf comments after here.'
    comment = indicatortools.LilyPondComment(string, 'after')
    attach(comment, note)
    string = 'More comments after.'
    comment = indicatortools.LilyPondComment(string, 'after')
    attach(comment, note)

    assert format(note) == stringtools.normalize(
        r'''
        \once \override Beam.thickness = #3
        c'8
        % Leaf comments after here.
        % More comments after.
        '''
        )

    assert inspect_(note).is_well_formed()



# -*- coding: utf-8 -*-
from __future__ import print_function
import inspect
import os
import sys
import traceback
from abjad.tools import systemtools
from abjad.tools.commandlinetools.ScorePackageScript import ScorePackageScript


class ManageMaterialScript(ScorePackageScript):
    '''
    Manages score package materials.

    ..  shell::

        ajv material --help

    '''

    ### CLASS VARIABLES ###

    alias = 'material'
    short_description = 'Manage score package materials.'

    ### PRIVATE METHODS ###

    def _collect_matching_paths(self, names):
        matching_paths = []
        valid_paths = self._list_material_subpackages()
        for name in names:
            if name.startswith('+'):
                name = '*{}*'.format('*'.join(name[1:]))
            for path in sorted(self._materials_path.glob(name)):
                if path not in matching_paths and path in valid_paths:
                    matching_paths.append(path)
        return matching_paths

    def _handle_create(self, material_name, force):
        print('Creating material subpackage {!r} ...'.format(material_name))
        target_path = self._name_to_score_subdirectory_path(
            material_name, 'materials', self._score_package_path)
        if target_path.exists() and not force:
            print('    Path exists: {}'.format(
                target_path.relative_to(self._score_package_path.parent)))
            sys.exit(1)
        metadata = self._read_score_metadata_json(self._score_package_path)
        metadata['score_package_name'] = self._score_package_path.name
        metadata['material_name'] = target_path.name
        source_name = 'example_material'
        source_path = self._get_boilerplate_path().joinpath(source_name)
        suffixes = ('.py', '.tex', '.ly', '.ily')
        for path in self._copy_tree(source_path, target_path):
            if path.is_file() and path.suffix in suffixes:
                self._template_file(path, **metadata)
        print('    Created {path!s}{sep}'.format(
            path=target_path.relative_to(self._score_repository_path),
            sep=os.path.sep))

    def _handle_edit(self, material_name):
        from abjad import abjad_configuration
        globbable_names = self._collect_globbable_names(material_name)
        print('Edit candidates: {!r} ...'.format(
            ' '.join(globbable_names)))
        matching_paths = self._collect_matching_paths(globbable_names)
        if not matching_paths:
            print('    No matching materials.')
            self._handle_list()
        command = [abjad_configuration.get_text_editor()]
        for path in matching_paths:
            command.append(str(path.joinpath('definition.py')))
        command = ' '.join(command)
        exit_code = self._call_subprocess(command)
        if exit_code:
            sys.exit(exit_code)

    def _handle_illustrate(self, material_name):
        globbable_names = self._collect_globbable_names(material_name)
        print('Illustration candidates: {!r} ...'.format(
            ' '.join(globbable_names)))
        matching_paths = self._collect_matching_paths(globbable_names)
        if not matching_paths:
            print('    No matching materials.')
            self._handle_list()
        for path in matching_paths:
            self._illustrate_one_material(
                material_directory_path=path
                )
            print('    Illustrated {path!s}{sep}'.format(
                path=path.relative_to(self._score_package_path.parent),
                sep=os.path.sep))
        for path in matching_paths:
            pdf_path = path.joinpath('illustration.pdf')
            systemtools.IOManager.open_file(str(pdf_path))

    def _handle_list(self):
        from abjad.tools import abctools
        basic_bases = (
            abctools.AbjadObject,
            abctools.AbjadValueObject,
            object,
            )
        print('Available materials:')
        all_materials = self._import_all_materials(verbose=False)
        if not all_materials:
            print('    No materials available.')
            sys.exit(2)
        materials = {}
        for material_name, material in all_materials.items():
            class_ = type(material)
            base = class_.__bases__[0]
            attrs = {attr.name: attr for attr in
                inspect.classify_class_attrs(class_)}
            if any(_ in class_.__bases__ for _ in basic_bases):
                base = class_
            elif getattr(class_, '__is_terminal_ajv_list_item__', False) and \
                attrs['__is_terminal_ajv_list_item__'].defining_class is class_:
                base = class_
            materials.setdefault(base, []).append((material_name, class_))
        #valid_paths = self._list_material_subpackages(self._score_package_path)
        materials = sorted(materials.items(), key=lambda pair: pair[0].__name__)
        for base, material_names in materials:
            print('    {}:'.format(base.__name__))
            for material_name, class_ in material_names:
                print('        {} [{}]'.format(material_name, class_.__name__))
        sys.exit(2)

    def _handle_render(self, material_name):
        globbable_names = self._collect_globbable_names(material_name)
        print('Rendering candidates: {!r} ...'.format(
            ' '.join(globbable_names)))
        matching_paths = self._collect_matching_paths(globbable_names)
        if not matching_paths:
            print('    No matching materials.')
            self._handle_list()
        for path in matching_paths:
            self._render_one_material(
                material_directory_path=path
                )
            print('    Rendered {path!s}{sep}'.format(
                path=path.relative_to(self._score_package_path.parent),
                sep=os.path.sep))
        for path in matching_paths:
            pdf_path = path.joinpath('illustration.pdf')
            systemtools.IOManager.open_file(str(pdf_path))

    def _illustrate_one_material(self, material_directory_path):
        print('Illustrating {path!s}{sep}'.format(
            path=material_directory_path.relative_to(self._score_package_path.parent),
            sep=os.path.sep))
        material = self._import_material(material_directory_path)
        if not hasattr(material, '__illustrate__'):
            template = '    Cannot illustrate material of type {}.'
            message = template.format(type(material).__name__)
            print(message)
            sys.exit(1)
        with systemtools.Timer() as timer:
            try:
                lilypond_file = material.__illustrate__()
            except:
                traceback.print_exc()
                sys.exit(1)
        self._report_time(timer, prefix='Abjad runtime')
        ly_path = self._write_lilypond_ly(
            lilypond_file=lilypond_file,
            output_directory_path=material_directory_path,
            )
        self._write_lilypond_pdf(
            ly_path=ly_path,
            output_directory_path=material_directory_path,
            )

    def _process_args(self, arguments):
        self._setup_paths(arguments.score_path)
        if arguments.edit is not None:
            self._handle_edit(material_name=arguments.edit)
        if arguments.illustrate is not None:
            self._handle_illustrate(material_name=arguments.illustrate)
        if arguments.list_:
            self._handle_list()
        if arguments.new:
            self._handle_create(force=arguments.force, material_name=arguments.new)
        if arguments.render is not None:
            self._handle_render(material_name=arguments.render)

    def _render_one_material(self, material_directory_path):
        print('Rendering {path!s}{sep}'.format(
            path=material_directory_path.relative_to(self._score_package_path.parent),
            sep=os.path.sep))
        ly_path = material_directory_path.joinpath('illustration.ly')
        if not ly_path.is_file():
            print('    illustration.ly is missing or malformed.')
            sys.exit(1)
        self._write_lilypond_pdf(
            ly_path=ly_path,
            output_directory_path=material_directory_path,
            )

    def _setup_argument_parser(self, parser):
        action_group = parser.add_argument_group('actions')
        action_group = action_group.add_mutually_exclusive_group(required=True)
        action_group.add_argument(
            '--new', '-N',
            help='create a new material',
            metavar='NAME',
            )
        action_group.add_argument(
            '--edit', '-E',
            help='edit materials',
            metavar='PATTERN',
            nargs='+',
            )
        action_group.add_argument(
            '--illustrate', '-I',
            help='illustrate materials',
            metavar='PATTERN',
            nargs='+',
            )
        action_group.add_argument(
            '--render', '-R',
            help='render material illustrations',
            metavar='PATTERN',
            nargs='+',
            )
        action_group.add_argument(
            '--list', '-L',
            dest='list_',
            help='list materials',
            action='store_true',
            )
#        action_group.add_argument(
#            '--copy', '-Y',
#            help='copy material',
#            metavar=('SOURCE', 'TARGET'),
#            nargs=2,
#            )
#        action_group.add_argument(
#            '--rename', '-M',
#            help='rename material',
#            metavar=('SOURCE', 'TARGET'),
#            nargs=2,
#            )
#        action_group.add_argument(
#            '--delete', '-D',
#            help='delete material',
#            metavar='NAME',
#            )
        common_group = parser.add_argument_group('common options')
        common_group.add_argument(
            '--score-path', '-s',
            metavar='SCORE',
            help='score path or package name',
            default=os.path.curdir,
            )
        common_group.add_argument(
            '--force', '-f',
            action='store_true',
            help='force overwriting',
            )



# -*- coding: utf-8 -*-


def offset_happens_before_timespan_stops(
    timespan=None,
    offset=None,
    hold=False,
    ):
    r'''Makes time relation indicating that `offset` happens
    before `timespan` stops.

    ::

        >>> relation = timespantools.offset_happens_before_timespan_stops()
        >>> print(format(relation))
        timespantools.OffsetTimespanTimeRelation(
            inequality=timespantools.CompoundInequality(
                [
                    timespantools.Inequality('offset < timespan.stop'),
                    ],
                logical_operator='and',
                ),
            )

    Returns time relation or boolean.
    '''
    from abjad.tools import timespantools

    inequality = timespantools.CompoundInequality([
        'offset < timespan.stop',
        ])

    time_relation = timespantools.OffsetTimespanTimeRelation(
        inequality,
        timespan=timespan,
        offset=offset)

    if time_relation.is_fully_loaded and not hold:
        return time_relation()
    else:
        return time_relation



# -*- coding: utf-8 -*-
import pytest
from abjad import *


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___01():
    r'''Override LilyPond Accidental grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    override(staff).accidental.color = 'red'

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff \with {
            \override Accidental.color = #red
        } {
            c'8
            d'8
            e'8
            f'8
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___02():
    r'''Override LilyPond Accidental grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    override(staff[1]).accidental.color = 'red'

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff {
            c'8
            \once \override Accidental.color = #red
            d'8
            e'8
            f'8
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___03():
    r'''Override LilyPond BarNumber grob.
    '''

    score = Score([Staff("c'8 d'8 e'8 f'8 g'8 a'8 b'8 c''8")])
    override(score).bar_number.break_visibility = \
        schemetools.Scheme('end-of-line-invisible')

    assert format(score) == stringtools.normalize(
        r'''
        \new Score \with {
            \override BarNumber.break-visibility = #end-of-line-invisible
        } <<
            \new Staff {
                c'8
                d'8
                e'8
                f'8
                g'8
                a'8
                b'8
                c''8
            }
        >>
        '''
        )
    assert inspect_(score).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___04():
    r'''Override LilyPond BarNumber grob.
    '''

    score = Score([Staff("c'8 d'8 e'8 f'8")])
    override(score).bar_number.color = 'red'

    assert format(score) == stringtools.normalize(
        r'''
        \new Score \with {
            \override BarNumber.color = #red
        } <<
            \new Staff {
                c'8
                d'8
                e'8
                f'8
            }
        >>
        '''
        )

    assert inspect_(score).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___05():
    r'''Override LilyPond Beam grob.
    '''

    voice = Voice("c'8 d'8 e'8 f'8")
    beam = Beam()
    attach(beam, voice[:])
    override(beam).beam.positions = (4, 4)

    assert format(voice) == stringtools.normalize(
        r'''
        \new Voice {
            \override Beam.positions = #'(4 . 4)
            c'8 [
            d'8
            e'8
            f'8 ]
            \revert Beam.positions
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___06():
    r'''Override LilyPond Clef grob.
    '''

    note = Note("c'4")
    override(note).clef.color = 'red'

    assert format(note) == stringtools.normalize(
        r'''
        \once \override Clef.color = #red
        c'4
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___07():
    r'''Override LilyPond Clef grob.
    '''

    note = Note("c'4")
    override(note).staff.clef.color = 'red'

    assert format(note) == stringtools.normalize(
        r'''
        \once \override Staff.Clef.color = #red
        c'4
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___08():
    r'''Override LilyPond Clef grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    override(staff).clef.color = 'red'

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff \with {
            \override Clef.color = #red
        } {
            c'8
            d'8
            e'8
            f'8
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___09():
    r'''Override LilyPond ClusterSpanner grob.
    '''

    cluster = scoretools.Cluster(Note(1, (1, 4)) * 4)
    override(cluster).cluster_spanner.style = 'ramp'
    override(cluster).cluster_spanner.padding = 0.1

    assert format(cluster) == stringtools.normalize(
        r'''
        \makeClusters {
            \override ClusterSpanner.padding = #0.1
            \override ClusterSpanner.style = #'ramp
            cs'4
            cs'4
            cs'4
            cs'4
            \revert ClusterSpanner.padding
            \revert ClusterSpanner.style
        }
        '''
        )

    del(override(cluster).cluster_spanner)

    assert format(cluster) == stringtools.normalize(
        r'''
        \makeClusters {
            cs'4
            cs'4
            cs'4
            cs'4
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___10():
    r'''Override LilyPond DynamicLineSpanner grob.
    '''

    voice = Voice("c'8 d'8 e'8 f'8")
    hairpin = Hairpin(descriptor='p < f')
    attach(hairpin, voice[:])
    override(hairpin).dynamic_line_spanner.staff_padding = 4

    assert format(voice) == stringtools.normalize(
        r'''
        \new Voice {
            \override DynamicLineSpanner.staff-padding = #4
            c'8 \< \p
            d'8
            e'8
            f'8 \f
            \revert DynamicLineSpanner.staff-padding
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___11():
    r'''Override LilyPond DynamicLineSpanner grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    override(staff).dynamic_line_spanner.staff_padding = 2
    override(staff).dynamic_line_spanner.Y_extent = (-1.5, 1.5)

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff \with {
            \override DynamicLineSpanner.Y-extent = #'(-1.5 . 1.5)
            \override DynamicLineSpanner.staff-padding = #2
        } {
            c'8
            d'8
            e'8
            f'8
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___12():
    r'''Override LilyPond DynamicText grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    beam = Beam()
    attach(beam, staff[:])
    dynamic = Dynamic('f')
    attach(dynamic, staff[0])
    spanner = spannertools.Spanner()
    override(spanner).dynamic_text.thickness = 3
    attach(spanner, staff[:])

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff {
            \override DynamicText.thickness = #3
            c'8 \f [
            d'8
            e'8
            f'8 ]
            \revert DynamicText.thickness
        }
        '''
        )

    assert inspect_(staff).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___13():
    r'''Override LilyPond DynamicText grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    override(staff).dynamic_text.staff_padding = 2
    override(staff).dynamic_text.Y_extent = (-1.5, 1.5)

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff \with {
            \override DynamicText.Y-extent = #'(-1.5 . 1.5)
            \override DynamicText.staff-padding = #2
        } {
            c'8
            d'8
            e'8
            f'8
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___14():
    r'''Override LilyPond DynamicTextSpanner grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    override(staff).dynamic_text_spanner.staff_padding = 2
    override(staff).dynamic_text_spanner.Y_extent = (-1.5, 1.5)

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff \with {
            \override DynamicTextSpanner.Y-extent = #'(-1.5 . 1.5)
            \override DynamicTextSpanner.staff-padding = #2
        } {
            c'8
            d'8
            e'8
            f'8
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___15():
    r'''Override LilyPond Glissando grob.
    '''

    voice = Voice("c'8 d'8 e'8 f'8")
    glissando = spannertools.Glissando()
    attach(glissando, voice[:])
    override(glissando).glissando.thickness = 3

    assert format(voice) == stringtools.normalize(
        r'''
        \new Voice {
            \override Glissando.thickness = #3
            c'8 \glissando
            d'8 \glissando
            e'8 \glissando
            f'8
            \revert Glissando.thickness
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___16():
    r'''Override LilyPond Hairpin grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    override(staff).hairpin.staff_padding = 2
    override(staff).hairpin.Y_extent = (-1.5, 1.5)

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff \with {
            \override Hairpin.Y-extent = #'(-1.5 . 1.5)
            \override Hairpin.staff-padding = #2
        } {
            c'8
            d'8
            e'8
            f'8
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___17():
    r'''Override LilyPond InstrumentName grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    set_(staff).instrument_name = markuptools.Markup(r'\circle { V }')
    override(staff).instrument_name.color = 'red'

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff \with {
            \override InstrumentName.color = #red
            instrumentName = \markup {
                \circle
                    {
                        V
                    }
                }
        } {
            c'8
            d'8
            e'8
            f'8
        }
        ''',
        )

    assert inspect_(staff).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___18():
    r'''Override LilyPond MetronomeMark grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    score = Score([staff])
    tempo = Tempo(Duration(1, 4), 58)
    attach(tempo, staff[0])
    override(score).metronome_mark.color = 'red'

    assert format(score) == stringtools.normalize(
        r'''
        \new Score \with {
            \override MetronomeMark.color = #red
        } <<
            \new Staff {
                \tempo 4=58
                c'8
                d'8
                e'8
                f'8
            }
        >>
        '''
        )

    assert inspect_(score).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___19():
    r'''Override LilyPond MultiMeasureRestGrob.
    '''

    staff = Staff([Note("c'4")])
    override(staff).multi_measure_rest.expand_limit = 12

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff \with {
            \override MultiMeasureRest.expand-limit = #12
        } {
            c'4
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___20():
    r'''Override LilyPond NonMusicalPaperColumn grob.
    '''

    score = Score([Staff("c'8 d'8 e'8 f'8")])
    override(score).non_musical_paper_column.line_break_permission = False
    override(score).non_musical_paper_column.page_break_permission = False

    assert format(score) == stringtools.normalize(
        r'''
        \new Score \with {
            \override NonMusicalPaperColumn.line-break-permission = ##f
            \override NonMusicalPaperColumn.page-break-permission = ##f
        } <<
            \new Staff {
                c'8
                d'8
                e'8
                f'8
            }
        >>
        '''
        )

    assert inspect_(score).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___21():
    r'''Override LilyPond NoteColumn grob.
    '''

    note = Note("c'4")
    override(note).note_column.ignore_collision = True

    assert format(note) == stringtools.normalize(
        r'''
        \once \override NoteColumn.ignore-collision = ##t
        c'4
        '''
        )

    assert inspect_(note).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___22():
    r'''Override LilyPond NoteColumn grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    override(staff).note_column.ignore_collision = True

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff \with {
            \override NoteColumn.ignore-collision = ##t
        } {
            c'8
            d'8
            e'8
            f'8
        }
        '''
        )

    assert inspect_(staff).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___23():
    r'''Override LilyPond NoteHead grob.
    '''

    note = Note(1, (1, 4))
    override(note).note_head.style = 'cross'

    assert override(note).note_head.style == 'cross'
    assert format(note) == stringtools.normalize(
        r'''
        \once \override NoteHead.style = #'cross
        cs'4
        '''
        )

    del(override(note).note_head.style)
    assert format(note.note_head) == "cs'"


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___24():
    r'''Notehead styles are handled just like all other grob overrides.
    '''

    note = Note(1, (1, 4))
    override(note).note_head.style = 'mystrangehead'

    assert override(note).note_head.style == 'mystrangehead'
    assert format(note) == stringtools.normalize(
        r'''
        \once \override NoteHead.style = #'mystrangehead
        cs'4
        '''
        )

    del(override(note).note_head.style)
    assert format(note.note_head) == "cs'"


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___25():
    r'''Notehead style overrides are handled just like all other
    note_head grob overrides, even for note_heads in chords.'''

    chord = Chord([1, 2, 3], (1, 4))
    chord.note_heads[0].tweak.style = 'harmonic'

    assert format(chord) == stringtools.normalize(
        r'''
        <
            \tweak style #'harmonic
            cs'
            d'
            ef'
        >4
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___26():
    r'''Notehead shape style overrides are just normal grob overrides.
    '''

    note = Note(1, (1, 4))
    override(note).note_head.style = 'triangle'

    assert override(note).note_head.style == 'triangle'
    assert format(note) == stringtools.normalize(
        r'''
        \once \override NoteHead.style = #'triangle
        cs'4
        '''
        )

    del(override(note).note_head.style)
    assert format(note) == "cs'4"


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___27():
    r'''Notehead solfege style overrides are just normal grob overrides.
    Modern versions of LilyPond now handles solfege overrides correctly.'''

    note = Note(1, (1, 4))
    override(note).note_head.style = 'do'

    assert override(note).note_head.style == 'do'
    assert format(note) == stringtools.normalize(
        r'''
        \once \override NoteHead.style = #'do
        cs'4
        '''
        )

    del(override(note).note_head.style)
    assert format(note) == "cs'4"


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___28():
    r'''Override LilyPond NoteHead grob.
    '''

    note = Note(13, (1, 4))
    override(note).note_head.transparent = True

    assert override(note).note_head.transparent
    assert format(note) == stringtools.normalize(
        r'''
        \once \override NoteHead.transparent = ##t
        cs''4
        '''
        )

    del(override(note).note_head.transparent)
    assert format(note) == "cs''4"


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___29():
    r'''Override LilyPond NoteHead grob.
    '''

    voice = Voice("c'8 d'8 e'8 f'8")
    override(voice).note_head.color = 'red'

    assert format(voice) == stringtools.normalize(
        r'''
        \new Voice \with {
            \override NoteHead.color = #red
        } {
            c'8
            d'8
            e'8
            f'8
        }
        '''
        )

    assert inspect_(voice).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___30():
    r'''Override LilyPond OctavationBracket grob.
    '''

    voice = Voice("c'8 d'8 e'8 f'8")
    octavation_spanner = spannertools.OctavationSpanner(start=1)
    attach(octavation_spanner, voice[:])
    override(octavation_spanner).staff.ottava_bracket.staff_position = 4

    assert format(voice) == stringtools.normalize(
        r'''
        \new Voice {
            \override Staff.OttavaBracket.staff-position = #4
            \ottava #1
            c'8
            d'8
            e'8
            f'8
            \ottava #0
            \revert Staff.OttavaBracket.staff-position
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___31():
    r'''Override LilyPond RehearsalMark grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    override(staff).rehearsal_mark.staff_padding = 2
    override(staff).rehearsal_mark.Y_extent = (-1.5, 1.5)

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff \with {
            \override RehearsalMark.Y-extent = #'(-1.5 . 1.5)
            \override RehearsalMark.staff-padding = #2
        } {
            c'8
            d'8
            e'8
            f'8
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___32():
    r'''Override LilyPond Rest grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    override(staff).rest.transparent = True

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff \with {
            \override Rest.transparent = ##t
        } {
            c'8
            d'8
            e'8
            f'8
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___33():
    r'''Override LilyPond Script grob.
    '''

    note = Note("c'4")
    articulation = Articulation('staccato')
    attach(articulation, note)
    override(note).script.color = 'red'

    assert format(note) == stringtools.normalize(
        r'''
        \once \override Script.color = #red
        c'4 -\staccato
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___34():
    r'''Override LilyPond Script grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    override(staff).script.staff_padding = 2
    override(staff).script.Y_extent = (-1.5, 1.5)

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff \with {
            \override Script.Y-extent = #'(-1.5 . 1.5)
            \override Script.staff-padding = #2
        } {
            c'8
            d'8
            e'8
            f'8
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___35():
    r'''Override LilyPond SpacingSpanner grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    beam = Beam()
    attach(beam, staff[:])
    override(beam).score.spacing_spanner.strict_grace_spacing = True
    override(beam).score.spacing_spanner.strict_note_spacing = True
    override(beam).score.spacing_spanner.uniform_stretching = True

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff {
            \override Score.SpacingSpanner.strict-grace-spacing = ##t
            \override Score.SpacingSpanner.strict-note-spacing = ##t
            \override Score.SpacingSpanner.uniform-stretching = ##t
            c'8 [
            d'8
            e'8
            f'8 ]
            \revert Score.SpacingSpanner.strict-grace-spacing
            \revert Score.SpacingSpanner.strict-note-spacing
            \revert Score.SpacingSpanner.uniform-stretching
        }
        '''
        )

    assert inspect_(staff).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___36():
    r'''Override LilyPond SpacingSpanner grob on Abjad containers.
    LilyPond SpacingSpanner lives at Score by default.
    Abjad SpacingSpanner overrides usually
    require context promotion.
    '''

    staff = Staff("abj: | 2/8 c'8 d'8 || 2/8 e'8 f'8 |")
    leaves = select(staff).by_leaf()
    beam = Beam()
    attach(beam, leaves)
    override(beam).score.spacing_spanner.strict_grace_spacing = True
    override(beam).score.spacing_spanner.strict_note_spacing = True
    override(beam).score.spacing_spanner.uniform_stretching = True

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff {
            {
                \time 2/8
                \override Score.SpacingSpanner.strict-grace-spacing = ##t
                \override Score.SpacingSpanner.strict-note-spacing = ##t
                \override Score.SpacingSpanner.uniform-stretching = ##t
                c'8 [
                d'8
            }
            {
                e'8
                f'8 ]
                \revert Score.SpacingSpanner.strict-grace-spacing
                \revert Score.SpacingSpanner.strict-note-spacing
                \revert Score.SpacingSpanner.uniform-stretching
            }
        }
        '''
        )

    assert inspect_(staff).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___37():
    r'''Override LilyPond SpacingSpanner grob.
    '''

    score = Score([])
    override(score).spacing_spanner.strict_grace_spacing = True
    override(score).spacing_spanner.strict_note_spacing = True
    override(score).spacing_spanner.uniform_stretching = True

    assert format(score) == stringtools.normalize(
        r'''
        \new Score \with {
            \override SpacingSpanner.strict-grace-spacing = ##t
            \override SpacingSpanner.strict-note-spacing = ##t
            \override SpacingSpanner.uniform-stretching = ##t
        } <<
        >>
        '''
        )

    assert not len(score)


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___38():
    r'''Override LilyPond SpanBar grob.
    '''

    score, treble, bass = scoretools.make_empty_piano_score()
    notes = [Note("c'8"), Note("d'8"), Note("e'8"), Note("f'8")]
    treble.extend(notes)
    notes = [Note("c'8"), Note("d'8"), Note("e'8"), Note("f'8")]
    bass.extend(notes)
    override(score).span_bar.color = 'red'

    assert format(score) == stringtools.normalize(
        r'''
        \new Score \with {
            \override SpanBar.color = #red
        } <<
            \new PianoStaff <<
                \context Staff = "treble" {
                    \clef "treble"
                    c'8
                    d'8
                    e'8
                    f'8
                }
                \context Staff = "bass" {
                    \clef "bass"
                    c'8
                    d'8
                    e'8
                    f'8
                }
            >>
        >>
        '''
        )

    assert inspect_(score).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___39():
    r'''Override LilyPond StaffSymbol grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    override(staff).staff_symbol.color = 'red'

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff \with {
            \override StaffSymbol.color = #red
        } {
            c'8
            d'8
            e'8
            f'8
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___40():
    r'''Override LilyPond StaffSymbol grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    override(staff[2]).staff.staff_symbol.color = 'red'

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff {
            c'8
            d'8
            \once \override Staff.StaffSymbol.color = #red
            e'8
            f'8
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___41():
    r'''Override LilyPond StaffSymbol grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    override(staff).staff_symbol.line_positions = \
        schemetools.SchemeVector(-4, -2, 2, 4)

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff \with {
            \override StaffSymbol.line-positions = #'(-4 -2 2 4)
        } {
            c'8
            d'8
            e'8
            f'8
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___42():
    r'''Override LilyPond Stem grob.
    '''

    note = Note(0, (1, 16))
    override(note).stem.stroke_style = \
        schemetools.Scheme('grace', force_quotes=True)

    assert format(note) == stringtools.normalize(
        r'''
        \once \override Stem.stroke-style = #"grace"
        c'16
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___43():
    r'''Override LilyPond StemTremolo grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    override(staff).stem_tremolo.slope = 0.5
    override(staff).stem_tremolo.staff_padding = 2

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff \with {
            \override StemTremolo.slope = #0.5
            \override StemTremolo.staff-padding = #2
        } {
            c'8
            d'8
            e'8
            f'8
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___44():
    r'''Override LilyPond SystemStartBar grob.
    '''

    score = Score([scoretools.StaffGroup([Staff("c'8 c'8 c'8 c'8 c'8 c'8 c'8 c'8")])])
    override(score).system_start_bar.collapse_height = 0
    override(score).system_start_bar.color = 'red'

    assert format(score) == stringtools.normalize(
        r'''
        \new Score \with {
            \override SystemStartBar.collapse-height = #0
            \override SystemStartBar.color = #red
        } <<
            \new StaffGroup <<
                \new Staff {
                    c'8
                    c'8
                    c'8
                    c'8
                    c'8
                    c'8
                    c'8
                    c'8
                }
            >>
        >>
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___45():
    r'''Override LilyPond TextScript grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    spanner = spannertools.Spanner()
    attach(spanner, staff[:])
    override(spanner).text_script.color = 'red'

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff {
            \override TextScript.color = #red
            c'8
            d'8
            e'8
            f'8
            \revert TextScript.color
        }
        '''
        )

    assert inspect_(staff).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___46():
    r'''Override LilyPond TextSpanner grob.
    '''

    staff = Staff("c'8 c'8 c'8 c'8")
    text_spanner = spannertools.TextSpanner()
    attach(text_spanner, staff[:])
    override(text_spanner).text_spanner.font_shape = 'italic'

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff {
            \override TextSpanner.font-shape = #'italic
            c'8 \startTextSpan
            c'8
            c'8
            c'8 \stopTextSpan
            \revert TextSpanner.font-shape
        }
        '''
        )

    assert inspect_(staff).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___47():
    r'''Override LilyPond Tie grob.
    '''

    note = Note("c'4")
    override(note).tie.color = 'red'

    assert format(note) == stringtools.normalize(
        r'''
        \once \override Tie.color = #red
        c'4
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___48():
    r'''Override LilyPond TimeSignature grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    override(staff).time_signature.transparent = True

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff \with {
            \override TimeSignature.transparent = ##t
        } {
            c'8
            d'8
            e'8
            f'8
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___49():
    r'''Override LilyPond TimeSignature grob.
    '''

    measure = Measure((4, 8), "c'8 d'8 e'8 f'8")
    override(measure).time_signature.transparent = True

    assert format(measure) == stringtools.normalize(
        r'''
        {
            \override TimeSignature.transparent = ##t
            \time 4/8
            c'8
            d'8
            e'8
            f'8
            \revert TimeSignature.transparent
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___50():
    r'''Override LilyPond TimeSignature grob.
    '''

    measure = Measure((4, 8), "c'8 d'8 e'8 f'8")
    override(measure).staff.time_signature.transparent = True

    assert format(measure) == stringtools.normalize(
        r'''
        {
            \override Staff.TimeSignature.transparent = ##t
            \time 4/8
            c'8
            d'8
            e'8
            f'8
            \revert Staff.TimeSignature.transparent
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___51():
    r'''Override LilyPond TrillPitchAccidental grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    override(staff).trill_pitch_accidental.staff_padding = 2
    override(staff).trill_pitch_accidental.Y_extent = (-1.5, 1.5)

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff \with {
            \override TrillPitchAccidental.Y-extent = #'(-1.5 . 1.5)
            \override TrillPitchAccidental.staff-padding = #2
        } {
            c'8
            d'8
            e'8
            f'8
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___52():
    r'''Override LilyPond TrillSpanner grob.
    '''

    voice = Voice("c'8 d'8 e'8 f'8")
    trill = spannertools.TrillSpanner()
    attach(trill, voice[:])
    override(trill).trill_spanner.color = 'red'

    assert format(voice) == stringtools.normalize(
        r'''
        \new Voice {
            \override TrillSpanner.color = #red
            c'8 \startTrillSpan
            d'8
            e'8
            f'8 \stopTrillSpan
            \revert TrillSpanner.color
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___53():
    r'''Override LilyPond TupletBracket grob.
    '''

    voice = Voice("c'8 d'8 e'8 f'8")
    beam = Beam()
    attach(beam, voice[:])
    override(voice).tuplet_bracket.direction = Down

    assert format(voice) == stringtools.normalize(
        r'''
        \new Voice \with {
            \override TupletBracket.direction = #down
        } {
            c'8 [
            d'8
            e'8
            f'8 ]
        }
        '''
        )

    assert inspect_(voice).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___54():
    r'''Override LilyPond TupletBracket grob.
    '''

    voice = Voice("c'8 d'8 e'8 f'8")
    beam = Beam()
    attach(beam, voice[:])
    override(voice[1]).tuplet_bracket.direction = Down

    assert format(voice) == stringtools.normalize(
        r'''
        \new Voice {
            c'8 [
            \once \override TupletBracket.direction = #down
            d'8
            e'8
            f'8 ]
        }
        '''
        )

    assert inspect_(voice).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___55():
    r'''Override LilyPond TupletNumber grob.
    '''

    tuplet = scoretools.FixedDurationTuplet(Duration(2, 8), "c'8 d'8 e'8")
    override(tuplet).tuplet_number.fraction = True

    assert format(tuplet) == stringtools.normalize(
        r'''
        \override TupletNumber.fraction = ##t
        \times 2/3 {
            c'8
            d'8
            e'8
        }
        \revert TupletNumber.fraction
        '''
        )

    assert inspect_(tuplet).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___56():
    r'''Override LilyPond TupletNumber grob.
    '''

    voice = Voice("c'8 d'8 e'8 f'8")
    beam = Beam()
    attach(beam, voice[:])
    override(voice).tuplet_number.fraction = True

    assert format(voice) == stringtools.normalize(
        r'''
        \new Voice \with {
            \override TupletNumber.fraction = ##t
        } {
            c'8 [
            d'8
            e'8
            f'8 ]
        }
        '''
        )

    assert inspect_(voice).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___57():
    r'''Override LilyPond TupletNumber grob.
    '''

    voice = Voice("c'8 d'8 e'8 f'8")
    beam = Beam()
    attach(beam, voice[:])
    override(voice[1]).tuplet_number.fraction = True

    assert format(voice) == stringtools.normalize(
        r'''
        \new Voice {
            c'8 [
            \once \override TupletNumber.fraction = ##t
            d'8
            e'8
            f'8 ]
        }
        '''
        )

    assert inspect_(voice).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___58():
    r'''Override LilyPond TupletNumber grob.
    '''

    voice = Voice("c'8 d'8 e'8 f'8")
    beam = Beam()
    attach(beam, voice[:])
    override(voice).tuplet_number.text = markuptools.Markup('6:4')

    assert format(voice) == stringtools.normalize(
        r'''
        \new Voice \with {
            \override TupletNumber.text = \markup { 6:4 }
        } {
            c'8 [
            d'8
            e'8
            f'8 ]
        }
        '''
        )

    assert inspect_(voice).is_well_formed()


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___59():
    r'''Override LilyPond VerticalAlignment grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    override(staff).vertical_alignment.staff_padding = 2
    override(staff).vertical_alignment.Y_extent = (-1.5, 1.5)

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff \with {
            \override VerticalAlignment.Y-extent = #'(-1.5 . 1.5)
            \override VerticalAlignment.staff-padding = #2
        } {
            c'8
            d'8
            e'8
            f'8
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___60():
    r'''Override LilyPond VerticalAxis grob.
    '''

    staff = Staff("c'8 d'8 e'8 f'8")
    override(staff).vertical_axis_group.staff_padding = 2
    override(staff).vertical_axis_group.Y_extent = (-1.5, 1.5)

    assert format(staff) == stringtools.normalize(
        r'''
        \new Staff \with {
            \override VerticalAxisGroup.Y-extent = #'(-1.5 . 1.5)
            \override VerticalAxisGroup.staff-padding = #2
        } {
            c'8
            d'8
            e'8
            f'8
        }
        '''
        )


def test_lilypondproxytools_LilyPondGrobNameManager___setattr___61():
    r'''InputSetExpression attribute on erroneous grob name raises exception.
    '''

    note = Note("c'8")
    assert pytest.raises(Exception, 'override(note).foo = True')



# -*- coding: utf-8 -*-
import os
import platform
import shutil
import unittest
from abjad.tools import abjadbooktools
from abjad.tools import stringtools


@unittest.skipIf(
    platform.python_implementation() != 'CPython',
    'Only for CPython.',
    )
class TestLaTeXDocumentHandler(unittest.TestCase):

    maxDiff = None

    test_directory = os.path.dirname(__file__)
    assets_directory = os.path.join(test_directory, 'assets')
    source_valid_path = os.path.join(test_directory, 'source_valid.tex')
    with open(source_valid_path, 'r') as file_pointer:
        source_valid_contents = file_pointer.read().rstrip()
    target_valid_path = os.path.join(test_directory, 'target_valid.tex')

    expected_valid_path = os.path.join(test_directory, 'expected_valid.tex')
    with open(expected_valid_path, 'r') as file_pointer:
        expected_valid_contents = file_pointer.read().rstrip()
    expected_asset_names = (
        'graphviz-31410f5aefd17473e91ebc219ddff36e.dot',
        'graphviz-31410f5aefd17473e91ebc219ddff36e.pdf',
        'lilypond-0b731cedacea34e85fbb92b66b42b40b.ly',
        'lilypond-0b731cedacea34e85fbb92b66b42b40b.pdf',
        'lilypond-0fa8f7f49d6d92ffcb7dd54b5b8e851c.ly',
        'lilypond-0fa8f7f49d6d92ffcb7dd54b5b8e851c.pdf',
        'lilypond-412bd86b762d452fb787992613b02ae1.ly',
        'lilypond-412bd86b762d452fb787992613b02ae1.pdf',
        'lilypond-6737d707e144fd1f0af98dd8007ebb4b.ly',
        'lilypond-6737d707e144fd1f0af98dd8007ebb4b.pdf',
        'lilypond-8812fc9449c0c47bf22dbf11bdfaeb7b.ly',
        'lilypond-8812fc9449c0c47bf22dbf11bdfaeb7b.pdf',
        )

    stylesheet_path = os.path.join(test_directory, 'stylesheet.ily')
    expected_styled_path = os.path.join(test_directory, 'expected_styled.tex')
    with open(expected_styled_path, 'r') as file_pointer:
        expected_styled_contents = file_pointer.read().rstrip()
    styled_asset_names = (
        'graphviz-31410f5aefd17473e91ebc219ddff36e.dot',
        'graphviz-31410f5aefd17473e91ebc219ddff36e.pdf',
        'lilypond-29a694fcaa88ad2e66126db254f6e44c.ly',
        'lilypond-29a694fcaa88ad2e66126db254f6e44c.pdf',
        'lilypond-60a8aa65d1dc86d7b3b3396d1dd25fa6.ly',
        'lilypond-60a8aa65d1dc86d7b3b3396d1dd25fa6.pdf',
        'lilypond-65404525b3a554691b49772ace1d914c.ly',
        'lilypond-65404525b3a554691b49772ace1d914c.pdf',
        'lilypond-8d473a765e713b5eb5d0a95f35161666.ly',
        'lilypond-8d473a765e713b5eb5d0a95f35161666.pdf',
        'lilypond-ef7d9d5db77fdbbf21122f8479e9e338.ly',
        'lilypond-ef7d9d5db77fdbbf21122f8479e9e338.pdf',
        )

    configuration_path = os.path.join(test_directory, 'configuration.cfg')
    expected_configured_path = os.path.join(
        test_directory, 'expected_configured.tex')
    with open(expected_configured_path, 'r') as file_pointer:
        expected_configured_contents = file_pointer.read().rstrip()

    def setUp(self):
        if os.path.exists(self.assets_directory):
            shutil.rmtree(self.assets_directory)
        if os.path.exists(self.target_valid_path):
            os.remove(self.target_valid_path)

    def tearDown(self):
        if os.path.exists(self.assets_directory):
            shutil.rmtree(self.assets_directory)
        if os.path.exists(self.target_valid_path):
            os.remove(self.target_valid_path)
        with open(self.source_valid_path, 'w') as file_pointer:
            file_pointer.write(self.source_valid_contents)

    def test_clean(self):
        document_handler = abjadbooktools.LaTeXDocumentHandler.from_path(
            input_file_path=self.source_valid_path)
        document_handler()
        with open(self.source_valid_path, 'r') as file_pointer:
            target_valid_contents = file_pointer.read()
        assert stringtools.normalize(str(target_valid_contents)) == \
            stringtools.normalize(str(self.expected_valid_contents))
        document_handler = abjadbooktools.LaTeXDocumentHandler.from_path(
            input_file_path=self.source_valid_path)
        document_handler(clean=True)
        with open(self.source_valid_path, 'r') as file_pointer:
            target_valid_contents = file_pointer.read()
        assert target_valid_contents == self.source_valid_contents

    def test_configured(self):
        configuration = abjadbooktools.AbjadBookScript()._read_config(
            self.configuration_path)
        assert not os.path.exists(self.target_valid_path)
        assert not os.path.exists(self.assets_directory)
        document_handler = abjadbooktools.LaTeXDocumentHandler.from_path(
            input_file_path=self.source_valid_path)
        document_handler(
            configuration=configuration,
            output_file_path=self.target_valid_path,
            )
        assert os.path.exists(self.target_valid_path)
        assert os.path.exists(self.assets_directory)
        with open(self.target_valid_path, 'r') as file_pointer:
            target_valid_contents = file_pointer.read()
        assert target_valid_contents == self.expected_configured_contents
        assert tuple(sorted(os.listdir(self.assets_directory))) == \
            self.expected_asset_names

    def test_double_source(self):
        assert not os.path.exists(self.target_valid_path)
        assert not os.path.exists(self.assets_directory)
        document_handler = abjadbooktools.LaTeXDocumentHandler.from_path(
            input_file_path=self.source_valid_path)
        document_handler(output_file_path=self.target_valid_path)
        assert os.path.exists(self.target_valid_path)
        assert os.path.exists(self.assets_directory)
        with open(self.target_valid_path, 'r') as file_pointer:
            target_valid_contents = file_pointer.read()
        assert target_valid_contents == self.expected_valid_contents
        assert tuple(sorted(os.listdir(self.assets_directory))) == \
            self.expected_asset_names

    def test_return_source(self):
        assert not os.path.exists(self.assets_directory)
        document_handler = abjadbooktools.LaTeXDocumentHandler.from_path(
            input_file_path=self.source_valid_path)
        target_valid_contents = document_handler(return_source=True)
        assert target_valid_contents == self.expected_valid_contents
        #assert not os.path.exists(self.assets_directory)

    def test_single_source(self):
        assert not os.path.exists(self.assets_directory)
        document_handler = abjadbooktools.LaTeXDocumentHandler.from_path(
            input_file_path=self.source_valid_path)
        document_handler()
        assert os.path.exists(self.source_valid_path)
        assert os.path.exists(self.assets_directory)
        with open(self.source_valid_path, 'r') as file_pointer:
            target_valid_contents = file_pointer.read()
        assert target_valid_contents == self.expected_valid_contents
        assert tuple(sorted(os.listdir(self.assets_directory))) == \
            self.expected_asset_names

    def test_single_source_rebuild(self):
        assert not os.path.exists(self.assets_directory)
        document_handler = abjadbooktools.LaTeXDocumentHandler.from_path(
            input_file_path=self.source_valid_path)
        document_handler()
        assert os.path.exists(self.source_valid_path)
        assert os.path.exists(self.assets_directory)
        with open(self.source_valid_path, 'r') as file_pointer:
            target_valid_contents = file_pointer.read()
        assert target_valid_contents == self.expected_valid_contents
        assert tuple(sorted(os.listdir(self.assets_directory))) == \
            self.expected_asset_names
        document_handler = abjadbooktools.LaTeXDocumentHandler.from_path(
            input_file_path=self.source_valid_path)
        document_handler()
        assert os.path.exists(self.source_valid_path)
        assert os.path.exists(self.assets_directory)
        with open(self.source_valid_path, 'r') as file_pointer:
            target_valid_contents = file_pointer.read()
        assert target_valid_contents == self.expected_valid_contents
        assert tuple(sorted(os.listdir(self.assets_directory))) == \
            self.expected_asset_names

    def test_skip_rendering(self):
        assert not os.path.exists(self.target_valid_path)
        assert not os.path.exists(self.assets_directory)
        document_handler = abjadbooktools.LaTeXDocumentHandler.from_path(
            input_file_path=self.source_valid_path)
        document_handler(
            output_file_path=self.target_valid_path,
            skip_rendering=True,
            )
        assert os.path.exists(self.target_valid_path)
        assert not os.path.exists(self.assets_directory)
        with open(self.target_valid_path, 'r') as file_pointer:
            target_valid_contents = file_pointer.read()
        assert target_valid_contents == self.expected_valid_contents

    @unittest.skipIf(
        platform.system() == 'Windows',
        'Windows path-handling woes.'
        )
    def test_stylesheet(self):
        assert not os.path.exists(self.target_valid_path)
        assert not os.path.exists(self.assets_directory)
        document_handler = abjadbooktools.LaTeXDocumentHandler.from_path(
            input_file_path=self.source_valid_path)
        document_handler(
            output_file_path=self.target_valid_path,
            stylesheet=self.stylesheet_path,
            )
        assert os.path.exists(self.target_valid_path)
        assert os.path.exists(self.assets_directory)
        with open(self.target_valid_path, 'r') as file_pointer:
            target_valid_contents = file_pointer.read()
        assert target_valid_contents == self.expected_styled_contents
        assert tuple(sorted(os.listdir(self.assets_directory))) == \
            self.styled_asset_names



# -*- coding: utf-8 -*-
import os
import platform
from abjad.tools import commandlinetools
from abjad.tools import systemtools
from base import ScorePackageScriptTestCase


class Test(ScorePackageScriptTestCase):

    expected_files = [
        'test_score/test_score/build/.gitignore',
        'test_score/test_score/build/assets/.gitignore',
        'test_score/test_score/build/assets/instrumentation.tex',
        'test_score/test_score/build/assets/performance-notes.tex',
        'test_score/test_score/build/letter-portrait/back-cover.tex',
        'test_score/test_score/build/letter-portrait/front-cover.tex',
        'test_score/test_score/build/letter-portrait/music.ly',
        'test_score/test_score/build/letter-portrait/parts.ly',
        'test_score/test_score/build/letter-portrait/preface.tex',
        'test_score/test_score/build/letter-portrait/score.tex',
        'test_score/test_score/build/parts.ily',
        'test_score/test_score/build/segments.ily',
        'test_score/test_score/build/segments/.gitignore',
        ]

    if platform.system().lower() == 'windows':
        expected_files = [_.replace('/', os.path.sep) for _ in expected_files]

    def test_exists(self):
        self.create_score()
        script = commandlinetools.ManageBuildTargetScript()
        command = ['--new']
        with systemtools.TemporaryDirectoryChange(str(self.score_path)):
            try:
                script(command)
            except SystemExit:
                raise RuntimeError('SystemExit')
            with systemtools.RedirectedStreams(stdout=self.string_io):
                with self.assertRaises(SystemExit) as context_manager:
                    script(command)
                assert context_manager.exception.code == 1
        self.compare_captured_output(r'''
        Creating build target 'letter-portrait' (8.5in x 11.0in)
            Path exists: test_score/build/letter-portrait
        '''.replace('/', os.path.sep))

    def test_explicit(self):
        self.create_score()
        script = commandlinetools.ManageBuildTargetScript()
        command = [
            '--new',
            '--paper-size', 'a3',
            '--orientation', 'landscape',
            ]
        with systemtools.RedirectedStreams(stdout=self.string_io):
            with systemtools.TemporaryDirectoryChange(str(self.score_path)):
                try:
                    script(command)
                except SystemExit:
                    raise RuntimeError('SystemExit')
        self.compare_captured_output(r'''
        Creating build target 'a3-landscape' (297mm x 420mm)
            Reading test_score/metadata.json ... OK!
            Created test_score/build/a3-landscape
        '''.replace('/', os.path.sep))
        path = self.build_path.joinpath('a3-landscape', 'score.tex')
        self.compare_lilypond_contents(path, r'''
            \documentclass{article}
            \usepackage[papersize={420mm, 297mm}]{geometry}
            \usepackage{pdfpages}
            \begin{document}

            \includepdf[pages=-]{front-cover.pdf}
            \includepdf[pages=-]{preface.pdf}
            \includepdf[angle=-90,pages=-]{music.pdf}
            \includepdf[pages=-]{back-cover.pdf}

            \end{document}
        ''')

    def test_force_replace(self):
        self.create_score()
        script = commandlinetools.ManageBuildTargetScript()
        command = ['-f', '--new']
        with systemtools.TemporaryDirectoryChange(str(self.score_path)):
            try:
                script(command)
            except SystemExit:
                raise RuntimeError('SystemExit')
            with systemtools.RedirectedStreams(stdout=self.string_io):
                try:
                    script(command)
                except SystemExit:
                    raise RuntimeError('SystemExit')
        self.compare_captured_output(r'''
        Creating build target 'letter-portrait' (8.5in x 11.0in)
            Reading test_score/metadata.json ... OK!
            Created test_score/build/letter-portrait
        '''.replace('/', os.path.sep))

    def test_implicit(self):
        self.create_score()
        script = commandlinetools.ManageBuildTargetScript()
        command = ['--new']
        with systemtools.RedirectedStreams(stdout=self.string_io):
            with systemtools.TemporaryDirectoryChange(str(self.score_path)):
                try:
                    script(command)
                except SystemExit:
                    raise RuntimeError('SystemExit')
        self.compare_captured_output(r'''
        Creating build target 'letter-portrait' (8.5in x 11.0in)
            Reading test_score/metadata.json ... OK!
            Created test_score/build/letter-portrait
        '''.replace('/', os.path.sep))
        self.compare_path_contents(self.build_path, self.expected_files)
        path = self.build_path.joinpath('letter-portrait', 'music.ly')
        self.compare_lilypond_contents(path, r'''
            \language "english"

            #(ly:set-option 'relative-includes #t)
            \include "../../stylesheets/stylesheet.ily"

            #(set-default-paper-size "letter" 'portrait)
            #(set-global-staff-size 12)

            \layout {
            }

            \paper {
            }

            \score {
                \include "../segments.ily"
            }
        ''')
        path = self.build_path.joinpath('letter-portrait', 'parts.ly')
        self.compare_lilypond_contents(path, r'''
            \language "english"

            #(ly:set-option 'relative-includes #t)
            \include "../../stylesheets/stylesheet.ily"
            \include "../../stylesheets/parts.ily"

            #(set-default-paper-size "letter" 'portrait)
            #(set-global-staff-size 12)

            \include "../parts.ily"
        ''')
        path = self.build_path.joinpath('letter-portrait', 'score.tex')
        self.compare_lilypond_contents(path, r'''
            \documentclass{article}
            \usepackage[papersize={8.5in, 11.0in}]{geometry}
            \usepackage{pdfpages}
            \begin{document}

            \includepdf[pages=-]{front-cover.pdf}
            \includepdf[pages=-]{preface.pdf}
            \includepdf[angle=0,pages=-]{music.pdf}
            \includepdf[pages=-]{back-cover.pdf}

            \end{document}
        ''')
        path = self.build_path.joinpath('letter-portrait', 'front-cover.tex')
        self.compare_lilypond_contents(path, r"""
        \documentclass[11pt]{report}
        <BLANKLINE>
        \usepackage[T1]{fontenc}
        \usepackage[papersize={8.5in, 11.0in}]{geometry}
        \usepackage{tikz}
        \usepackage{xltxtra,fontspec,xunicode}
        \usetikzlibrary{calc}
        <BLANKLINE>
        %\defaultfontfeatures{Scale=MatchLowercase}
        %\setromanfont[Numbers=Uppercase]{Didot}
        \newlength{\drop}
        \parindent=0pt
        <BLANKLINE>
        \begin{document}
            \begin{titlepage}
        <BLANKLINE>
                \begin{tikzpicture}[remember picture, overlay]
                    \draw[line width = 1.6pt]
                        ($(current page.north west) + (1in,-1in)$)
                        rectangle
                        ($(current page.south east) + (-1in,1in)$);
                    \draw[line width = 0.4pt]
                        ($(current page.north west) + (1in,-1in) + (2.8pt,-2.8pt)$)
                        rectangle
                        ($(current page.south east) + (-1in,1in) + (-2.8pt,2.8pt)$);
                \end{tikzpicture}
        <BLANKLINE>
                \drop=0.1\textheight
                \centering
        <BLANKLINE>
                \vspace*{2\baselineskip}
        <BLANKLINE>
                \rule{\textwidth}{1.6pt}\vspace*{-\baselineskip}\vspace*{2pt}
                \rule{\textwidth}{0.4pt}\\[\baselineskip]
                {
                    \fontsize{6cm}{1em}\selectfont
                    TEST SCORE
                }
                \rule{\textwidth}{0.4pt}\vspace*{-\baselineskip}\vspace{3.2pt}
                \rule{\textwidth}{1.6pt}\\[\baselineskip]
        <BLANKLINE>
                \vspace*{2\baselineskip}
        <BLANKLINE>
                {
                    \itshape
                    Composed by
                }
                \\
                {
                    \Large
                    JOSIAH WOLF OBERHOLTZER
                    \par
                }
        <BLANKLINE>
                \vspace*{2\baselineskip}
        <BLANKLINE>
                {
                    \Large
                    2016
                }
        <BLANKLINE>
                \vfill
        <BLANKLINE>
            \end{titlepage}
        \end{document}
        """)
        path = self.build_path.joinpath('letter-portrait', 'back-cover.tex')
        self.compare_lilypond_contents(path, r'''
        \documentclass[11pt]{report}

        \usepackage[utf8]{inputenc}
        \usepackage{eurosym}
        \usepackage[papersize={8.5in, 11.0in}]{geometry}
        \usepackage{tikz}
        \usepackage{xltxtra,fontspec,xunicode}
        \usetikzlibrary{calc}

        %\defaultfontfeatures{Scale=MatchLowercase}
        %\setromanfont[Numbers=Uppercase]{Didot}

        \begin{document}
            \begin{titlepage}

                \begin{tikzpicture}[remember picture, overlay]
                    \draw[line width = 1.6pt]
                        ($(current page.north west) + (1in,-1in)$)
                        rectangle
                        ($(current page.south east) + (-1in,1in)$);
                    \draw[line width = 0.4pt]
                        ($(current page.north west) + (1in,-1in) + (2.8pt,-2.8pt)$)
                        rectangle
                        ($(current page.south east) + (-1in,1in) + (-2.8pt,2.8pt)$);
                \end{tikzpicture}

                \vfill

                \begin{center}

                \vspace{1em}
                {
                    \itshape
                    Scores available from the composer at\\
                }
                {
                    www.josiahwolfoberholtzer.com\\
                }
                \vspace{1em}
                \euro 750 / \$1000

                \end{center}

            \end{titlepage}
        \end{document}
        ''')

    def test_internal_path(self):
        self.create_score()
        script = commandlinetools.ManageBuildTargetScript()
        command = ['--new']
        internal_path = self.score_path.joinpath('test_score', 'build')
        assert internal_path.exists()
        with systemtools.RedirectedStreams(stdout=self.string_io):
            with systemtools.TemporaryDirectoryChange(str(internal_path)):
                try:
                    script(command)
                except SystemExit:
                    raise RuntimeError('SystemExit')
        self.compare_captured_output(r'''
        Creating build target 'letter-portrait' (8.5in x 11.0in)
            Reading test_score/metadata.json ... OK!
            Created test_score/build/letter-portrait
        '''.replace('/', os.path.sep))

    def test_named(self):
        self.create_score()
        script = commandlinetools.ManageBuildTargetScript()
        command = [
            '--new',
            'World Premiere Version',
            '--paper-size', 'a3',
            '--orientation', 'landscape',
            ]
        with systemtools.RedirectedStreams(stdout=self.string_io):
            with systemtools.TemporaryDirectoryChange(str(self.score_path)):
                try:
                    script(command)
                except SystemExit:
                    raise RuntimeError('SystemExit')
        self.compare_captured_output(r'''
        Creating build target 'world-premiere-version' (297mm x 420mm)
            Reading test_score/metadata.json ... OK!
            Created test_score/build/world-premiere-version
        '''.replace('/', os.path.sep))
